import inspect
from typing import Dict, Any, Optional
from abc import ABC


class NotionBlockElement(ABC):
    """Base class for elements that can be converted between Markdown and Notion."""

    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown to Notion block."""

    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion block to markdown."""

    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if this element can handle the given markdown text."""
        return bool(NotionBlockElement.markdown_to_notion(text))

    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if this element can handle the given Notion block."""
        return bool(NotionBlockElement.notion_to_markdown(block))

    @staticmethod
    def is_multiline() -> bool:
        return False

    @classmethod
    def get_llm_documentation(cls) -> str:
        """
        Returns documentation specifically formatted for LLM system prompts.
        Can be overridden by subclasses to provide custom LLM-friendly documentation.

        By default, returns the class docstring.
        """


@classmethod
def get_llm_prompt_content(cls) -> dict:
    """
    Returns a dictionary with information for LLM prompts about this element.
    This default implementation extracts information from the class docstring.
    Subclasses should override this method to provide more structured information.

    Returns:
        Dictionary with documentation information
    """
    return {"description": inspect.cleandoc(cls.__doc__ or ""), "examples": []}



import re
from typing import Dict, Any, Optional, List, Tuple
from typing_extensions import override

from notionary.core.converters.elements.notion_block_element import NotionBlockElement


class QuoteElement(NotionBlockElement):
    """Class for converting between Markdown blockquotes and Notion quote blocks with background color support."""

    # Mapping von Markdown-Farbnamen zu Notion-Farbnamen
    COLOR_MAPPING = {
        "gray": "gray_background",
        "brown": "brown_background",
        "orange": "orange_background",
        "yellow": "yellow_background",
        "green": "green_background",
        "blue": "blue_background",
        "purple": "purple_background",
        "pink": "pink_background",
        "red": "red_background",
    }

    # Umgekehrtes Mapping für die Rückkonvertierung
    REVERSE_COLOR_MAPPING = {v: k for k, v in COLOR_MAPPING.items()}

    @staticmethod
    def find_matches(text: str) -> List[Tuple[int, int, Dict[str, Any]]]:
        """
        Find all blockquote matches in the text and return their positions and blocks.

        Args:
            text: The input markdown text

        Returns:
            List of tuples (start_pos, end_pos, block)
        """
        quote_pattern = re.compile(r"^\s*>\s?(.*)", re.MULTILINE)
        matches = []

        # Find all potential quote line matches
        quote_matches = list(quote_pattern.finditer(text))
        if not quote_matches:
            return []

        # Group consecutive quote lines
        i = 0
        while i < len(quote_matches):
            start_match = quote_matches[i]
            start_pos = start_match.start()

            # Find consecutive quote lines
            j = i + 1
            while j < len(quote_matches):
                # Check if this is the next line (considering newlines)
                if (
                    text[quote_matches[j - 1].end() : quote_matches[j].start()].count(
                        "\n"
                    )
                    == 1
                    or
                    # Or if it's an empty line followed by a quote line
                    (
                        text[
                            quote_matches[j - 1].end() : quote_matches[j].start()
                        ].strip()
                        == ""
                        and text[
                            quote_matches[j - 1].end() : quote_matches[j].start()
                        ].count("\n")
                        <= 2
                    )
                ):
                    j += 1
                else:
                    break

            end_pos = quote_matches[j - 1].end()
            quote_text = text[start_pos:end_pos]

            # Create the block
            block = QuoteElement.markdown_to_notion(quote_text)
            if block:
                matches.append((start_pos, end_pos, block))

            i = j

        return matches

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown blockquote to Notion block with background color support."""
        if not text:
            return None

        quote_pattern = re.compile(r"^\s*>\s?(.*)", re.MULTILINE)

        # Check if it's a blockquote
        if not quote_pattern.search(text):
            return None

        # Extract quote content
        lines = text.split("\n")
        quote_lines = []
        color = "default"  # Standardfarbe

        # Überprüfen, ob der erste Nicht-Leerzeichen-Inhalt eine Farbangabe ist
        first_line = None
        for line in lines:
            quote_match = quote_pattern.match(line)
            if quote_match and quote_match.group(1).strip():
                first_line = quote_match.group(1).strip()
                break

        # Farbangabe in eckigen Klammern prüfen
        if first_line:
            color_match = re.match(r"^\[background:(\w+)\]\s*(.*)", first_line)
            if color_match:
                potential_color = color_match.group(1).lower()
                if potential_color in QuoteElement.COLOR_MAPPING:
                    color = QuoteElement.COLOR_MAPPING[potential_color]
                    # Erste Zeile ohne Farbangabe neu hinzufügen
                    first_line = color_match.group(2)

        # Inhalte extrahieren
        processing_first_color_line = True
        for line in lines:
            quote_match = quote_pattern.match(line)
            if quote_match:
                content = quote_match.group(1)
                # Farbangabe in der ersten Zeile entfernen
                if (
                    processing_first_color_line
                    and content.strip()
                    and re.match(r"^\[background:(\w+)\]", content.strip())
                ):
                    content = re.sub(r"^\[background:(\w+)\]\s*", "", content)
                    processing_first_color_line = False
                quote_lines.append(content)
            elif not line.strip() and quote_lines:
                # Allow empty lines within the quote
                quote_lines.append("")

        if not quote_lines:
            return None

        quote_content = "\n".join(quote_lines).strip()

        # Create rich_text elements directly
        rich_text = [{"type": "text", "text": {"content": quote_content}}]

        return {"type": "quote", "quote": {"rich_text": rich_text, "color": color}}

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion quote block to markdown with background color support."""
        if block.get("type") != "quote":
            return None

        rich_text = block.get("quote", {}).get("rich_text", [])
        color = block.get("quote", {}).get("color", "default")

        # Extract the text content
        content = QuoteElement._extract_text_content(rich_text)

        # Format as markdown blockquote
        lines = content.split("\n")
        formatted_lines = []

        # Füge die Farbinformation zur ersten Zeile hinzu, falls nicht default
        if color != "default" and color in QuoteElement.REVERSE_COLOR_MAPPING:
            markdown_color = QuoteElement.REVERSE_COLOR_MAPPING.get(color)
            first_line = lines[0] if lines else ""
            formatted_lines.append(f"> [background:{markdown_color}] {first_line}")
            lines = lines[1:] if len(lines) > 1 else []

        # Füge die restlichen Zeilen hinzu
        for line in lines:
            formatted_lines.append(f"> {line}")

        return "\n".join(formatted_lines)

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if this element can handle the given markdown text."""
        quote_pattern = re.compile(r"^\s*>\s?(.*)", re.MULTILINE)
        return bool(quote_pattern.search(text))

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if this element can handle the given Notion block."""
        return block.get("type") == "quote"

    @override
    @staticmethod
    def is_multiline() -> bool:
        """Blockquotes can span multiple lines."""
        return True

    @staticmethod
    def _extract_text_content(rich_text: List[Dict[str, Any]]) -> str:
        """Extract plain text content from Notion rich_text elements."""
        result = ""
        for text_obj in rich_text:
            if text_obj.get("type") == "text":
                result += text_obj.get("text", {}).get("content", "")
            elif "plain_text" in text_obj:
                result += text_obj.get("plain_text", "")
        return result

    @override
    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """Returns information for LLM prompts about this element."""
        return {
            "description": "Creates blockquotes that visually distinguish quoted text with optional background colors.",
            "when_to_use": "Use blockquotes for quoting external sources, highlighting important statements, or creating visual emphasis for key information.",
            "syntax": [
                "> Text - Simple blockquote",
                "> [background:color] Text - Blockquote with colored background",
            ],
            "color_options": [
                "gray",
                "brown",
                "orange",
                "yellow",
                "green",
                "blue",
                "purple",
                "pink",
                "red",
            ],
            "examples": [
                "> This is a simple blockquote without any color",
                "> [background:blue] This is a blockquote with blue background",
                "> Multi-line quotes\n> continue like this\n> across several lines",
            ],
        }



from typing import Dict, Any, List, Optional, Tuple

from notionary.core.converters.registry.block_element_registry import (
    BlockElementRegistry,
)
from notionary.core.converters.registry.block_element_registry_builder import (
    BlockElementRegistryBuilder,
)


class MarkdownToNotionConverter:
    SPACER_MARKER = "<!-- spacer -->"
    MULTILINE_CONTENT_MARKER = "<!-- REMOVED_MULTILINE_CONTENT -->"
    TOGGLE_MARKER = "<!-- toggle_content -->"

    def __init__(self, block_registry: Optional[BlockElementRegistry] = None):
        """
        Initialize the MarkdownToNotionConverter.

        Args:
            block_registry: Optional registry of Notion block elements
        """
        self._block_registry = (
            block_registry or BlockElementRegistryBuilder().create_standard_registry()
        )

        self._setup_element_callbacks()

    def _setup_element_callbacks(self) -> None:
        """Registriert den Converter als Callback für Elemente, die ihn benötigen."""

        for element in self._block_registry.get_elements():
            if hasattr(element, "set_converter_callback"):
                element.set_converter_callback(self.convert)

    def convert(self, markdown_text: str) -> List[Dict[str, Any]]:
        """
        Convert markdown text to Notion API block format.

        Args:
            markdown_text: The markdown text to convert

        Returns:
            List of Notion blocks
        """
        if not markdown_text:
            return []

        # Process toggles first
        processed_text, toggle_blocks = self._extract_toggle_elements(markdown_text)

        # Process other multiline elements
        processed_text, multiline_blocks = self._extract_multiline_elements(
            processed_text
        )

        # Process remaining text line by line
        line_blocks = self._process_text_lines(processed_text)

        # Combine and sort all blocks
        all_blocks = toggle_blocks + multiline_blocks + line_blocks
        all_blocks.sort(key=lambda x: x[0])

        # Extract just the blocks from position tuples
        blocks = [block for _, _, block in all_blocks]

        # Process spacing between blocks
        return self._process_block_spacing(blocks)

    def _extract_toggle_elements(
        self, text: str
    ) -> Tuple[str, List[Tuple[int, int, Dict[str, Any]]]]:
        """
        Extract toggle elements and their nested content using the ToggleElement class.

        Args:
            text: The text to process

        Returns:
            Tuple of (processed text, list of (start_pos, end_pos, block) tuples)
        """
        # Find toggle element in registry
        toggle_element = None
        for element in self._block_registry.get_elements():
            if (
                element.is_multiline()
                and hasattr(element, "match_markdown")
                and element.__name__ == "ToggleElement"
            ):
                toggle_element = element
                break

        if not toggle_element:
            # No toggle element found, return text as is
            return text, []

        # Use the find_matches method of ToggleElement to find and process all toggles
        # Pass the converter's convert method as a callback to process nested content
        toggle_blocks = toggle_element.find_matches(text, self.convert)

        if not toggle_blocks:
            return text, []

        # Create a processed text with toggle markers
        lines = text.split("\n")
        processed_lines = lines.copy()

        # Replace toggle content with markers
        for start_pos, end_pos, _ in reversed(toggle_blocks):
            # Calculate line indices for this toggle
            start_line_index = 0
            current_pos = 0
            for i, line in enumerate(lines):
                line_length = len(line) + 1  # +1 for newline
                if current_pos <= start_pos < current_pos + line_length:
                    start_line_index = i
                    break
                current_pos += line_length

            end_line_index = start_line_index
            current_pos = 0
            for i, line in enumerate(lines):
                line_length = len(line) + 1  # +1 for newline
                if current_pos <= end_pos < current_pos + line_length:
                    end_line_index = i
                    break
                current_pos += line_length

            # Replace toggle content with markers
            num_lines = end_line_index - start_line_index + 1
            for i in range(start_line_index, start_line_index + num_lines):
                processed_lines[i] = self.TOGGLE_MARKER

        processed_text = "\n".join(processed_lines)
        return processed_text, toggle_blocks

    def _extract_multiline_elements(
        self, text: str
    ) -> Tuple[str, List[Tuple[int, int, Dict[str, Any]]]]:
        """
        Extract multiline elements and remove them from the text.

        Args:
            text: The text to process

        Returns:
            Tuple of (processed text, list of (start_pos, end_pos, block) tuples)
        """
        if not text:
            return text, []

        multiline_blocks = []
        processed_text = text

        # Get all multiline elements except ToggleElement
        multiline_elements = [
            element
            for element in self._block_registry.get_multiline_elements()
            if element.__name__ != "ToggleElement"
        ]

        if not multiline_elements:
            return text, []

        for element in multiline_elements:
            if not hasattr(element, "find_matches"):
                continue

            # Find all matches for this element (pass the convert method as callback if needed)
            if hasattr(element, "set_converter_callback"):
                matches = element.find_matches(processed_text, self.convert)
            else:
                matches = element.find_matches(processed_text)

            if not matches:
                continue

            multiline_blocks.extend(matches)

            # Remove matched content from the text to avoid processing it again
            processed_text = self._replace_matched_content_with_markers(
                processed_text, matches
            )

        return processed_text, multiline_blocks

    def _replace_matched_content_with_markers(
        self, text: str, matches: List[Tuple[int, int, Dict[str, Any]]]
    ) -> str:
        """Replace matched content with marker placeholders to preserve line structure."""
        for start, end, _ in reversed(matches):
            num_newlines = text[start:end].count("\n")
            text = (
                text[:start]
                + "\n"
                + self.MULTILINE_CONTENT_MARKER
                + "\n" * num_newlines
                + text[end:]
            )
        return text

    def _process_text_lines(self, text: str) -> List[Tuple[int, int, Dict[str, Any]]]:
        """
        Process text line by line for single-line elements.

        Args:
            text: The text to process

        Returns:
            List of (start_pos, end_pos, block) tuples
        """
        if not text:
            return []

        line_blocks = []
        lines = text.split("\n")

        current_pos = 0
        current_paragraph = []
        paragraph_start = 0
        in_todo_sequence = False

        for line in lines:
            line_length = len(line) + 1  # +1 for newline

            # Skip marker lines
            if self._is_marker_line(line):
                current_pos += line_length
                continue

            # Check for spacer marker
            if self._is_spacer_marker(line):
                line_blocks.append(
                    (
                        current_pos,
                        current_pos + line_length,
                        self._create_empty_paragraph(),
                    )
                )
                current_pos += line_length
                continue

            # Process todos first to keep them grouped
            todo_block = self._extract_todo_item(line)
            if todo_block:
                self._handle_todo_item(
                    todo_block,
                    line_length,
                    current_pos,
                    current_paragraph,
                    paragraph_start,
                    line_blocks,
                    in_todo_sequence,
                )
                in_todo_sequence = True
                current_pos += line_length
                continue

            if in_todo_sequence:
                in_todo_sequence = False

            if not line.strip():
                self._process_paragraph_if_present(
                    current_paragraph, paragraph_start, current_pos, line_blocks
                )
                current_paragraph = []
                current_pos += line_length
                continue

            special_block = self._extract_special_block(line)
            if special_block:
                self._process_paragraph_if_present(
                    current_paragraph, paragraph_start, current_pos, line_blocks
                )
                line_blocks.append(
                    (current_pos, current_pos + line_length, special_block)
                )
                current_paragraph = []
                current_pos += line_length
                continue

            # Handle as part of paragraph
            if not current_paragraph:
                paragraph_start = current_pos
            current_paragraph.append(line)
            current_pos += line_length

        # Process any remaining paragraph content
        self._process_paragraph_if_present(
            current_paragraph, paragraph_start, current_pos, line_blocks
        )

        return line_blocks

    def _is_marker_line(self, line: str) -> bool:
        """Check if a line is any kind of marker line that should be skipped."""
        return self._is_multiline_marker(line) or self._is_toggle_marker(line)

    def _is_multiline_marker(self, line: str) -> bool:
        """Check if a line is a multiline content marker."""
        return line.strip() == self.MULTILINE_CONTENT_MARKER

    def _is_toggle_marker(self, line: str) -> bool:
        """Check if a line is a toggle content marker."""
        return line.strip() == self.TOGGLE_MARKER

    def _is_spacer_marker(self, line: str) -> bool:
        """Check if a line is a spacer marker."""
        return line.strip() == self.SPACER_MARKER

    def _extract_todo_item(self, line: str) -> Optional[Dict[str, Any]]:
        """
        Try to extract a todo item from a line.

        Returns:
            Todo block if line is a todo item, None otherwise
        """
        for element in self._block_registry.get_elements():
            if (
                not element.is_multiline()
                and hasattr(element, "match_markdown")
                and element.__name__ == "TodoElement"
                and element.match_markdown(line)
            ):
                return element.markdown_to_notion(line)
        return None

    def _handle_todo_item(
        self,
        todo_block: Dict[str, Any],
        line_length: int,
        current_pos: int,
        current_paragraph: List[str],
        paragraph_start: int,
        line_blocks: List[Tuple[int, int, Dict[str, Any]]],
        in_todo_sequence: bool,
    ) -> None:
        """Handle a todo item line."""
        # If we were building a paragraph, finish it before starting todos
        if not in_todo_sequence and current_paragraph:
            self._process_paragraph_if_present(
                current_paragraph, paragraph_start, current_pos, line_blocks
            )
            current_paragraph.clear()

        line_blocks.append((current_pos, current_pos + line_length, todo_block))

    def _extract_special_block(self, line: str) -> Optional[Dict[str, Any]]:
        """
        Try to extract a special block (not paragraph) from a line.

        Returns:
            Block if line is a special block, None otherwise
        """
        for element in self._block_registry.get_elements():
            if (
                not element.is_multiline()
                and hasattr(element, "match_markdown")
                and element.match_markdown(line)
            ):
                block = element.markdown_to_notion(line)
                if block and block.get("type") != "paragraph":
                    return block
        return None

    def _process_paragraph_if_present(
        self,
        paragraph_lines: List[str],
        start_pos: int,
        end_pos: int,
        blocks: List[Tuple[int, int, Dict[str, Any]]],
    ) -> None:
        """
        Process a paragraph and add it to the blocks list if valid.

        Args:
            paragraph_lines: Lines that make up the paragraph
            start_pos: Starting position of the paragraph
            end_pos: Ending position of the paragraph
            blocks: List to add the processed paragraph block to
        """
        if not paragraph_lines:
            return

        paragraph_text = "\n".join(paragraph_lines)
        block = self._block_registry.markdown_to_notion(paragraph_text)

        if not block:
            return

        blocks.append((start_pos, end_pos, block))

    def _process_block_spacing(
        self, blocks: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Process blocks and add spacing only where no explicit spacer is present.

        Args:
            blocks: List of Notion blocks

        Returns:
            List of Notion blocks with processed spacing
        """
        if not blocks:
            return blocks

        final_blocks = []
        i = 0

        while i < len(blocks):
            current_block = blocks[i]
            final_blocks.append(current_block)

            # Check if this is a multiline element that needs spacing
            if not self._is_multiline_block_type(current_block.get("type")):
                i += 1
                continue

            # Check if the next block is already a spacer
            if i + 1 < len(blocks) and self._is_empty_paragraph(blocks[i + 1]):
                # Next block is already a spacer, don't add another
                pass
            else:
                # No explicit spacer found, add one automatically
                final_blocks.append(self._create_empty_paragraph())

            i += 1

        return final_blocks

    def _is_multiline_block_type(self, block_type: str) -> bool:
        """
        Check if a block type corresponds to a multiline element.

        Args:
            block_type: The type of block to check

        Returns:
            True if the block type is a multiline element, False otherwise
        """
        if not block_type:
            return False

        multiline_elements = self._block_registry.get_multiline_elements()

        for element in multiline_elements:
            element_name = element.__name__.lower()
            if block_type in element_name:
                return True

            if hasattr(element, "match_notion"):
                dummy_block = {"type": block_type}
                if element.match_notion(dummy_block):
                    return True

        return False

    def _is_empty_paragraph(self, block: Dict[str, Any]) -> bool:
        """
        Check if a block is an empty paragraph.

        Args:
            block: The block to check

        Returns:
            True if it's an empty paragraph, False otherwise
        """
        if block.get("type") != "paragraph":
            return False

        rich_text = block.get("paragraph", {}).get("rich_text", [])
        return not rich_text or len(rich_text) == 0

    def _create_empty_paragraph(self) -> Dict[str, Any]:
        """
        Create an empty paragraph block.

        Returns:
            Empty paragraph block dictionary
        """
        return {"type": "paragraph", "paragraph": {"rich_text": []}}



from typing import Dict, List
from notionary.core.database.notion_database_schema import NotionDatabaseSchema


class DatabaseSchemaService:
    """Service für den Zugriff auf Datenbankschema-Informationen"""

    def __init__(self, schema: NotionDatabaseSchema):
        self._schema = schema

    async def get_property_types(self) -> Dict[str, str]:
        """
        Get all property types for the database.

        Returns:
            Dictionary mapping property names to their types
        """
        return await self._schema.get_property_types()

    async def get_select_options(self, property_name: str) -> List[Dict[str, str]]:
        """
        Get options for a select, multi-select, or status property.

        Args:
            property_name: Name of the property

        Returns:
            List of select options with name, id, and color (if available)
        """
        options = await self._schema.get_select_options(property_name)
        return [
            {
                "name": option.get("name", ""),
                "id": option.get("id", ""),
                "color": option.get("color", ""),
            }
            for option in options
        ]

    async def get_relation_options(
        self, property_name: str, limit: int = 100
    ) -> List[Dict[str, str]]:
        """
        Get available options for a relation property.

        Args:
            property_name: Name of the relation property
            limit: Maximum number of options to retrieve

        Returns:
            List of relation options with id and title
        """
        options = await self._schema.get_relation_options(property_name, limit)
        return [
            {"id": option.get("id", ""), "title": option.get("title", "")}
            for option in options
        ]



from typing import Any, Dict, List, Optional
from notionary.core.converters.registry.block_element_registry import (
    BlockElementRegistry,
)
from notionary.core.converters.registry.block_element_registry_builder import (
    BlockElementRegistryBuilder,
)
from notionary.core.notion_client import NotionClient
from notionary.core.page.page_content_manager import PageContentManager
from notionary.util.logging_mixin import LoggingMixin
from notionary.core.page.meta_data.metadata_editor import MetadataEditor
from notionary.util.uuid_utils import extract_uuid, format_uuid, is_valid_uuid


class NotionPageManager(LoggingMixin):
    """
    High-Level Fassade zur Verwaltung von Inhalten und Metadaten einer Notion-Seite.
    """

    def __init__(
        self,
        page_id: Optional[str] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        token: Optional[str] = None,
    ):
        if not page_id and not url:
            raise ValueError("Either page_id or url must be provided")

        if not page_id and url:
            page_id = extract_uuid(url)
            if not page_id:
                raise ValueError(f"Could not extract a valid UUID from the URL: {url}")

        page_id = format_uuid(page_id)
        if not page_id or not is_valid_uuid(page_id):
            raise ValueError(f"Invalid UUID format: {page_id}")

        self._page_id = page_id
        self.url = url
        self._title = title

        self._client = NotionClient(token=token)

        self._block_element_registry = (
            BlockElementRegistryBuilder.create_standard_registry()
        )

        self._page_content_manager = PageContentManager(
            page_id=page_id,
            client=self._client,
            block_registry=self._block_element_registry,
        )
        self._metadata = MetadataEditor(page_id, self._client)

    @property
    def page_id(self) -> Optional[str]:
        """Get the title of the page."""
        return self._page_id

    @property
    def title(self) -> Optional[str]:
        return self._title

    @property
    def block_registry(self) -> BlockElementRegistry:
        return self._block_element_registry

    @block_registry.setter
    def block_registry(self, block_registry: BlockElementRegistry) -> None:
        """Set the block element registry for the page content manager."""

        self._block_element_registry = block_registry

        self._page_content_manager = PageContentManager(
            page_id=self._page_id, client=self._client, block_registry=block_registry
        )

    async def append_markdown(self, markdown: str) -> str:
        return await self._page_content_manager.append_markdown(markdown)

    async def clear(self) -> str:
        return await self._page_content_manager.clear()

    async def replace_content(self, markdown: str) -> str:
        await self._page_content_manager.clear()
        return await self._page_content_manager.append_markdown(markdown)

    async def get_blocks(self) -> List[Dict[str, Any]]:
        return await self._page_content_manager.get_blocks()

    async def get_block_children(self, block_id: str) -> List[Dict[str, Any]]:
        return await self._page_content_manager.get_block_children(block_id)

    async def get_page_blocks_with_children(self) -> List[Dict[str, Any]]:
        return await self._page_content_manager.get_page_blocks_with_children()

    async def get_text(self) -> str:
        return await self._page_content_manager.get_text()

    async def set_title(self, title: str) -> Optional[Dict[str, Any]]:
        return await self._metadata.set_title(title)

    async def set_page_icon(
        self, emoji: Optional[str] = None, external_url: Optional[str] = None
    ) -> Optional[Dict[str, Any]]:
        return await self._metadata.set_icon(emoji, external_url)

    async def set_page_cover(self, external_url: str) -> Optional[Dict[str, Any]]:
        return await self._metadata.set_cover(external_url)



from typing import Dict, Any, Optional
from typing_extensions import override
import re

from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter
from notionary.core.converters.elements.notion_block_element import NotionBlockElement


class CalloutElement(NotionBlockElement):
    """
    Handles conversion between Markdown callouts and Notion callout blocks.

    Markdown callout syntax:
    - !> [emoji] Text - Callout with custom emoji
    - !> {color} [emoji] Text - Callout with custom color and emoji
    - !> Text - Simple callout with default emoji and color

    Where:
    - {color} can be one of Notion's color options (e.g., "blue_background")
    - [emoji] is any emoji character
    - Text is the callout content with optional inline formatting
    """

    COLOR_PATTERN = r"(?:(?:{([a-z_]+)})?\s*)?"
    EMOJI_PATTERN = r"(?:\[([^\]]+)\])?\s*"
    TEXT_PATTERN = r"(.+)"

    # Combine the patterns
    PATTERN = re.compile(
        r"^!>\s+"  # Callout prefix
        + COLOR_PATTERN
        + EMOJI_PATTERN
        + TEXT_PATTERN
        + r"$"  # End of line
    )

    DEFAULT_EMOJI = "💡"
    DEFAULT_COLOR = "gray_background"

    VALID_COLORS = [
        "default",
        "gray",
        "brown",
        "orange",
        "yellow",
        "green",
        "blue",
        "purple",
        "pink",
        "red",
        "gray_background",
        "brown_background",
        "orange_background",
        "yellow_background",
        "green_background",
        "blue_background",
        "purple_background",
        "pink_background",
        "red_background",
    ]

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text is a markdown callout."""
        return text.strip().startswith("!>") and bool(
            CalloutElement.PATTERN.match(text)
        )

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion callout."""
        return block.get("type") == "callout"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown callout to Notion callout block."""
        callout_match = CalloutElement.PATTERN.match(text)
        if not callout_match:
            return None

        color = callout_match.group(1)
        emoji = callout_match.group(2)
        content = callout_match.group(3)

        if not emoji:
            emoji = CalloutElement.DEFAULT_EMOJI

        if not color or color not in CalloutElement.VALID_COLORS:
            color = CalloutElement.DEFAULT_COLOR

        return {
            "type": "callout",
            "callout": {
                "rich_text": TextInlineFormatter.parse_inline_formatting(content),
                "icon": {"type": "emoji", "emoji": emoji},
                "color": color,
            },
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion callout block to markdown callout."""
        if block.get("type") != "callout":
            return None

        callout_data = block.get("callout", {})
        rich_text = callout_data.get("rich_text", [])
        icon = callout_data.get("icon", {})
        color = callout_data.get("color", CalloutElement.DEFAULT_COLOR)

        text = TextInlineFormatter.extract_text_with_formatting(rich_text)
        if not text:
            return None

        emoji = ""
        if icon and icon.get("type") == "emoji":
            emoji = icon.get("emoji", "")

        color_str = ""
        if color and color != CalloutElement.DEFAULT_COLOR:
            color_str = f"{{{color}}} "

        emoji_str = ""
        if emoji:
            emoji_str = f"[{emoji}] "

        return f"!> {color_str}{emoji_str}{text}"

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """
        Returns a dictionary with all information needed for LLM prompts about this element.
        Includes description, usage guidance, syntax options, and examples.
        """
        return {
            "description": "Creates a callout block to highlight important information with an icon and background color.",
            "when_to_use": "Use callouts when you want to draw attention to important information, tips, warnings, or notes that stand out from the main content.",
            "syntax": [
                "!> Text - Simple callout with default emoji (💡) and color (gray background)",
                "!> [emoji] Text - Callout with custom emoji",
                "!> {color} [emoji] Text - Callout with custom color and emoji",
            ],
            "color_options": [
                "default",
                "gray",
                "brown",
                "orange",
                "yellow",
                "green",
                "blue",
                "purple",
                "pink",
                "red",
                "gray_background",
                "brown_background",
                "orange_background",
                "yellow_background",
                "green_background",
                "blue_background",
                "purple_background",
                "pink_background",
                "red_background",
            ],
            "examples": [
                "!> This is a default callout with the light bulb emoji",
                "!> [🔔] This is a callout with a bell emoji",
                "!> {blue_background} [💧] This is a blue callout with a water drop emoji",
                "!> {yellow_background} [⚠️] Warning: This is an important note to pay attention to",
            ],
        }



from typing import Any, Dict, List, Optional, Union, TypedDict, cast
from notionary.core.database.notion_database_schema import NotionDatabaseSchema
from notionary.core.notion_client import NotionClient
from notionary.core.page.property_formatter import NotionPropertyFormatter
from notionary.util.logging_mixin import LoggingMixin


class NotionRelationItem(TypedDict):
    id: str


class NotionRelationProperty(TypedDict):
    relation: List[NotionRelationItem]


class NotionFormattedValue(TypedDict, total=False):
    title: List[Dict[str, Any]]
    rich_text: List[Dict[str, Any]]
    select: Dict[str, str]
    multi_select: List[Dict[str, str]]
    relation: List[NotionRelationItem]
    number: Union[int, float]
    date: Dict[str, str]
    checkbox: bool
    url: str
    email: str
    phone_number: str


class PageCreationResponse(TypedDict):
    id: str
    parent: Dict[str, str]
    properties: Dict[str, Any]


class NotionRelationHandler(LoggingMixin):
    """
    Handler for managing relations in Notion databases.
    Provides a unified interface for working with relations.
    """

    def __init__(self, client: NotionClient, db_schema: NotionDatabaseSchema) -> None:
        self._client = client
        self._db_schema = db_schema
        self._formatter = NotionPropertyFormatter()

    async def find_relation_by_title(
        self, database_id: str, relation_prop_name: str, title: str
    ) -> Optional[str]:
        """
        Finds a relation ID based on the title of the entry in the target database.
        """
        target_db_id = await self._db_schema.get_relation_database_id(
            relation_prop_name
        )
        if not target_db_id:
            self.logger.error(
                "No target database found for relation '%s' in database %s",
                relation_prop_name,
                database_id,
            )
            return None

        options = await self._db_schema.get_relation_options(relation_prop_name)

        for option in options:
            if option["title"].lower() == title.lower():
                self.logger.debug("Relation entry '%s' found: %s", title, option["id"])
                return option["id"]

        self.logger.warning("Relation entry '%s' not found", title)
        return None

    async def _get_title_properties(
        self, database_id: str, title: str
    ) -> Optional[Dict[str, NotionFormattedValue]]:
        """
        Determines the title property for a database and formats the value.
        """
        if not await self._db_schema.load():
            self.logger.error("Could not load database schema for %s", database_id)
            return None

        property_types = await self._db_schema.get_property_types()

        title_prop_name: Optional[str] = None
        for name, prop_type in property_types.items():
            if prop_type == "title":
                title_prop_name = name
                break

        if not title_prop_name:
            self.logger.error("No title property found in database %s", database_id)
            return None

        formatted_title = self._formatter.format_value("title", title)
        if not formatted_title:
            self.logger.error("Could not format title '%s'", title)
            return None

        return {title_prop_name: cast(NotionFormattedValue, formatted_title)}


class DatabaseWritter(LoggingMixin):
    """
    Enhanced class for creating and updating pages in Notion databases.
    Supports both simple properties and relations.
    """

    def __init__(
        self, client: NotionClient, db_schema: Optional[NotionDatabaseSchema] = None
    ) -> None:
        """
        Initialize with a NotionClient and optionally a NotionDatabaseSchema.

        Args:
            client: The Notion API client
            db_schema: Optional database schema instance
        """
        self._client = client
        self._formatter = NotionPropertyFormatter()

        self._active_schema: Optional[NotionDatabaseSchema] = db_schema
        self._relation_handler: Optional[NotionRelationHandler] = None

        if db_schema:
            self._relation_handler = NotionRelationHandler(client, db_schema)

    async def _ensure_schema_for_database(self, database_id: str) -> bool:
        """
        Stellt sicher, dass ein Schema für die angegebene Datenbank geladen ist.

        Args:
            database_id: ID der Datenbank

        Returns:
            True, wenn das Schema erfolgreich geladen wurde
        """
        if self._active_schema and self._active_schema.database_id == database_id:
            return True

        self._active_schema = NotionDatabaseSchema(database_id, self._client)
        self._relation_handler = NotionRelationHandler(
            self._client, self._active_schema
        )

        return await self._active_schema.load()

    async def create_page(
        self,
        database_id: str,
        properties: Dict[str, Any],
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Optional[PageCreationResponse]:
        """
        Creates a new page in a database with support for relations.
        """
        # Stelle sicher, dass wir ein Schema für diese Datenbank haben
        if not await self._ensure_schema_for_database(database_id):
            self.logger.error("Could not load schema for database %s", database_id)
            return None

        formatted_props = await self._format_properties(database_id, properties)
        if not formatted_props:
            return None

        if relations:
            relation_props = await self._process_relations(database_id, relations)
            if relation_props:
                formatted_props.update(relation_props)

        data: Dict[str, Any] = {
            "parent": {"database_id": database_id},
            "properties": formatted_props,
        }

        result = await self._client.post("pages", data)
        if not result:
            self.logger.error("Error creating page in database %s", database_id)
            return None

        self.logger.info("Page successfully created in database %s", database_id)
        return cast(PageCreationResponse, result)

    async def update_page(
        self,
        page_id: str,
        properties: Optional[Dict[str, Any]] = None,
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Updates a page with support for relations.
        """
        page_data = await self._client.get(f"pages/{page_id}")
        if (
            not page_data
            or "parent" not in page_data
            or "database_id" not in page_data["parent"]
        ):
            self.logger.error("Could not determine database ID for page %s", page_id)
            return None

        database_id = page_data["parent"]["database_id"]

        # Stelle sicher, dass wir ein Schema für diese Datenbank haben
        if not await self._ensure_schema_for_database(database_id):
            self.logger.error("Could not load schema for database %s", database_id)
            return None

        if not properties and not relations:
            self.logger.warning("No properties or relations specified for update")
            return page_data

        update_props: Dict[str, NotionFormattedValue] = {}

        if properties:
            formatted_props = await self._format_properties(database_id, properties)
            if formatted_props:
                update_props.update(formatted_props)

        if relations:
            relation_props = await self._process_relations(database_id, relations)
            if relation_props:
                update_props.update(relation_props)

        if not update_props:
            self.logger.warning("No valid properties to update for page %s", page_id)
            return None

        data = {"properties": update_props}

        result = await self._client.patch(f"pages/{page_id}", data)
        if not result:
            self.logger.error("Error updating page %s", page_id)
            return None

        self.logger.info("Page %s successfully updated", page_id)
        return result

    async def delete_page(self, page_id: str) -> bool:
        """
        Deletes a page (archives it in Notion).
        """
        data = {"archived": True}

        result = await self._client.patch(f"pages/{page_id}", data)
        if not result:
            self.logger.error("Error deleting page %s", page_id)
            return False

        self.logger.info("Page %s successfully deleted (archived)", page_id)
        return True

    async def _format_properties(
        self, database_id: str, properties: Dict[str, Any]
    ) -> Optional[Dict[str, NotionFormattedValue]]:
        """
        Formats properties according to their types in the database.
        """
        if not self._active_schema:
            self.logger.error("No active schema available for database %s", database_id)
            return None

        property_types = await self._active_schema.get_property_types()
        if not property_types:
            self.logger.error(
                "Could not get property types for database %s", database_id
            )
            return None

        formatted_props: Dict[str, NotionFormattedValue] = {}

        for prop_name, value in properties.items():
            if prop_name not in property_types:
                self.logger.warning(
                    "Property '%s' does not exist in database %s",
                    prop_name,
                    database_id,
                )
                continue

            prop_type = property_types[prop_name]

            formatted_value = self._formatter.format_value(prop_type, value)
            if formatted_value:
                formatted_props[prop_name] = cast(NotionFormattedValue, formatted_value)
            else:
                self.logger.warning(
                    "Could not format value for property '%s' of type '%s'",
                    prop_name,
                    prop_type,
                )

        return formatted_props

    async def _process_relations(
        self, database_id: str, relations: Dict[str, Union[str, List[str]]]
    ) -> Dict[str, NotionRelationProperty]:
        """
        Processes relation properties and converts titles to IDs.
        """
        if not self._relation_handler:
            self.logger.error("No relation handler available")
            return {}

        formatted_relations: Dict[str, NotionRelationProperty] = {}
        property_types = (
            await self._active_schema.get_property_types()
            if self._active_schema
            else {}
        )

        for prop_name, titles in relations.items():
            relation_property = await self._process_single_relation(
                database_id, prop_name, titles, property_types
            )
            if relation_property:
                formatted_relations[prop_name] = relation_property

        return formatted_relations

    async def _process_single_relation(
        self,
        database_id: str,
        prop_name: str,
        titles: Union[str, List[str]],
        property_types: Dict[str, str],
    ) -> Optional[NotionRelationProperty]:
        """
        Process a single relation property and convert titles to IDs.

        Args:
            database_id: The database ID
            prop_name: The property name
            titles: The title or list of titles to convert
            property_types: Dictionary of property types

        Returns:
            A formatted relation property or None if invalid
        """
        if prop_name not in property_types:
            self.logger.warning(
                "Property '%s' does not exist in database %s", prop_name, database_id
            )
            return None

        prop_type = property_types[prop_name]
        if prop_type != "relation":
            self.logger.warning(
                "Property '%s' is not a relation (type: %s)", prop_name, prop_type
            )
            return None

        title_list: List[str] = [titles] if isinstance(titles, str) else titles
        relation_ids = await self._get_relation_ids(database_id, prop_name, title_list)

        if not relation_ids:
            return None

        return {"relation": [{"id": rel_id} for rel_id in relation_ids]}

    async def _get_relation_ids(
        self, database_id: str, prop_name: str, titles: List[str]
    ) -> List[str]:
        """
        Get relation IDs for a list of titles.

        Args:
            database_id: The database ID
            prop_name: The property name
            titles: List of titles to convert

        Returns:
            List of relation IDs
        """
        relation_ids: List[str] = []

        for title in titles:
            relation_id = await self._relation_handler.find_relation_by_title(
                database_id, prop_name, title
            )

            if relation_id:
                relation_ids.append(relation_id)
            else:
                self.logger.warning(
                    "Could not find relation ID for '%s' in '%s'", title, prop_name
                )

        return relation_ids



from typing import Dict, Optional, Any, List, Union, TYPE_CHECKING
from notionary.core.database.models.page_result import PageResult
from notionary.core.database.notion_database_schema import NotionDatabaseSchema
from notionary.core.database.notion_database_writer import DatabaseWritter
from notionary.core.notion_client import NotionClient

from notionary.core.page.notion_page_manager import NotionPageManager
from notionary.exceptions.database_exceptions import (
    PageNotFoundException,
    PageOperationError,
)
from notionary.exceptions.page_creation_exception import PageCreationException
from notionary.util.uuid_utils import format_uuid


class DatabasePageService:
    """Service für den Umgang mit Datenbankseiten"""

    def __init__(
        self,
        client: NotionClient,
        schema: NotionDatabaseSchema,
        writer: DatabaseWritter,
    ):
        self._client = client
        self._schema = schema
        self._writer = writer

    def _format_page_id(self, page_id: str) -> str:
        """
        Format a page ID to ensure it's in the correct format.

        Args:
            page_id: The page ID to format

        Returns:
            The formatted page ID
        """
        return format_uuid(page_id) or page_id

    async def create_page(
        self,
        database_id: str,
        properties: Dict[str, Any],
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> PageResult:
        """
        Create a new page in the database.

        Args:
            database_id: The database ID to create the page in
            properties: Dictionary of property names and values
            relations: Optional dictionary of relation property names and titles

        Returns:
            Result object with success status and page information
        """
        try:
            response = await self._writer.create_page(
                database_id, properties, relations
            )

            if not response:
                return {
                    "success": False,
                    "message": f"Failed to create page in database {database_id}",
                }

            page_id = response.get("id", "")
            page_url = response.get("url", None)

            return {"success": True, "page_id": page_id, "url": page_url}

        except PageCreationException as e:
            return {"success": False, "message": str(e)}

    async def update_page(
        self,
        page_id: str,
        properties: Optional[Dict[str, Any]] = None,
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> PageResult:
        """
        Update an existing page.

        Args:
            page_id: The ID of the page to update
            properties: Dictionary of property names and values to update
            relations: Optional dictionary of relation property names and titles

        Returns:
            Result object with success status and message
        """
        try:
            formatted_page_id = self._format_page_id(page_id)

            response = await self._writer.update_page(
                formatted_page_id, properties, relations
            )

            if not response:
                return {
                    "success": False,
                    "message": f"Failed to update page {formatted_page_id}",
                }

            return {"success": True, "page_id": formatted_page_id}

        except PageOperationError as e:
            return {"success": False, "message": f"Error: {str(e)}"}

    async def delete_page(self, page_id: str) -> PageResult:
        """
        Delete (archive) a page.

        Args:
            page_id: The ID of the page to delete

        Returns:
            Result object with success status and message
        """
        try:
            formatted_page_id = self._format_page_id(page_id)

            success = await self._writer.delete_page(formatted_page_id)

            if not success:
                return {
                    "success": False,
                    "message": f"Failed to delete page {formatted_page_id}",
                }

            return {"success": True, "page_id": formatted_page_id}

        except PageOperationError as e:
            return {"success": False, "message": f"Error: {str(e)}"}

    async def get_page_manager(self, page_id: str) -> Optional[NotionPageManager]:
        """
        Get a NotionPageManager for a specific page.

        Args:
            page_id: The ID of the page

        Returns:
            NotionPageManager instance or None if the page wasn't found
        """
        formatted_page_id = self._format_page_id(page_id)

        try:
            page_data = await self._client.get(f"pages/{formatted_page_id}")

            if not page_data:
                return None

            return NotionPageManager(
                page_id=formatted_page_id, url=page_data.get("url")
            )

        except PageNotFoundException:
            return None



from typing import Any, AsyncGenerator, Dict, List, Optional, Union

from notionary.core.database.database_info_service import DatabaseInfoService
from notionary.core.database.database_query_service import DatabaseQueryService
from notionary.core.database.database_schema_service import DatabaseSchemaService
from notionary.core.database.models.page_result import PageResult
from notionary.core.database.page_service import DatabasePageService
from notionary.core.notion_client import NotionClient
from notionary.core.database.notion_database_schema import NotionDatabaseSchema
from notionary.core.database.notion_database_writer import DatabaseWritter
from notionary.core.page.notion_page_manager import NotionPageManager
from notionary.exceptions.database_exceptions import (
    DatabaseInitializationError,
    PropertyError,
)
from notionary.util.logging_mixin import LoggingMixin
from notionary.util.uuid_utils import format_uuid


class NotionDatabaseManager(LoggingMixin):
    """
    High-level facade for working with Notion databases.
    Provides simplified operations for creating, reading, updating and deleting pages.

    Note:
        It is recommended to create instances of this class using the NotionDatabaseFactory
        instead of directly calling the constructor.
    """

    def __init__(self, database_id: str, token: Optional[str] = None):
        """
        Initialize the database facade with a database ID.

        Note:
            It's recommended to use NotionDatabaseFactory to create instances of this class
            rather than using this constructor directly.

        Args:
            database_id: The ID of the Notion database
            token: Optional Notion API token (uses environment variable if not provided)
        """
        self.database_id = format_uuid(database_id) or database_id
        self._client = NotionClient(token=token)
        self._schema = NotionDatabaseSchema(self.database_id, self._client)
        self._writer = DatabaseWritter(self._client, self._schema)
        self._initialized = False

        self._info_service = DatabaseInfoService(self._client, self.database_id)
        self._page_service = DatabasePageService(
            self._client, self._schema, self._writer
        )
        self._query_service = DatabaseQueryService(self._schema)
        self._schema_service = DatabaseSchemaService(self._schema)

    @property
    def title(self) -> Optional[str]:
        """Get the database title."""
        return self._info_service.title

    async def initialize(self) -> bool:
        """
        Initialize the database facade by loading the schema.

        This method needs to be called after creating a new instance via the constructor.
        When using NotionDatabaseFactory, this is called automatically.
        """
        try:
            success = await self._schema.load()
            if not success:
                self.logger.error(
                    "Failed to load schema for database %s", self.database_id
                )
                return False

            await self._info_service.load_title()
            self.logger.debug("Loaded database title: %s", self.title)

            self._initialized = True
            return True
        except Exception as e:
            self.logger.error("Error initializing database: %s", str(e))
            return False

    async def _ensure_initialized(self) -> None:
        """
        Ensure the database manager is initialized before use.

        Raises:
            DatabaseInitializationError: If the database isn't initialized
        """
        if not self._initialized:
            raise DatabaseInitializationError(
                self.database_id,
                "Database manager not initialized. Call initialize() first.",
            )

    async def get_database_name(self) -> Optional[str]:
        """
        Get the name of the current database.

        Returns:
            The database name or None if it couldn't be retrieved
        """
        await self._ensure_initialized()

        if self.title:
            return self.title

        try:
            return await self._info_service.load_title()
        except PropertyError as e:
            self.logger.error("Error getting database name: %s", str(e))
            return None

    async def get_property_types(self) -> Dict[str, str]:
        """
        Get all property types for the database.

        Returns:
            Dictionary mapping property names to their types
        """
        await self._ensure_initialized()
        return await self._schema_service.get_property_types()

    async def get_select_options(self, property_name: str) -> List[Dict[str, str]]:
        """
        Get options for a select, multi-select, or status property.

        Args:
            property_name: Name of the property

        Returns:
            List of select options with name, id, and color (if available)
        """
        await self._ensure_initialized()
        return await self._schema_service.get_select_options(property_name)

    async def get_relation_options(
        self, property_name: str, limit: int = 100
    ) -> List[Dict[str, str]]:
        """
        Get available options for a relation property.

        Args:
            property_name: Name of the relation property
            limit: Maximum number of options to retrieve

        Returns:
            List of relation options with id and title
        """
        await self._ensure_initialized()
        return await self._schema_service.get_relation_options(property_name, limit)

    async def create_page(
        self,
        properties: Dict[str, Any],
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> PageResult:
        """
        Create a new page in the database.

        Args:
            properties: Dictionary of property names and values
            relations: Optional dictionary of relation property names and titles

        Returns:
            Result object with success status and page information
        """
        await self._ensure_initialized()

        result = await self._page_service.create_page(
            self.database_id, properties, relations
        )

        if result["success"]:
            self.logger.info(
                "Created page %s in database %s",
                result.get("page_id", ""),
                self.database_id,
            )
        else:
            self.logger.warning("Page creation failed: %s", result.get("message", ""))

        return result

    async def update_page(
        self,
        page_id: str,
        properties: Optional[Dict[str, Any]] = None,
        relations: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> PageResult:
        """
        Update an existing page.

        Args:
            page_id: The ID of the page to update
            properties: Dictionary of property names and values to update
            relations: Optional dictionary of relation property names and titles

        Returns:
            Result object with success status and message
        """
        await self._ensure_initialized()

        self.logger.debug("Updating page %s", page_id)

        result = await self._page_service.update_page(page_id, properties, relations)

        if result["success"]:
            self.logger.info("Successfully updated page %s", result.get("page_id", ""))
        else:
            self.logger.error(
                "Error updating page %s: %s", page_id, result.get("message", "")
            )

        return result

    async def delete_page(self, page_id: str) -> PageResult:
        """
        Delete (archive) a page.

        Args:
            page_id: The ID of the page to delete

        Returns:
            Result object with success status and message
        """
        await self._ensure_initialized()

        self.logger.debug("Deleting page %s", page_id)

        result = await self._page_service.delete_page(page_id)

        if result["success"]:
            self.logger.info("Successfully deleted page %s", result.get("page_id", ""))
        else:
            self.logger.error(
                "Error deleting page %s: %s", page_id, result.get("message", "")
            )

        return result

    async def get_pages(
        self,
        limit: int = 100,
        filter_conditions: Optional[Dict[str, Any]] = None,
        sorts: Optional[List[Dict[str, Any]]] = None,
    ) -> List[NotionPageManager]:
        """
        Get all pages from the database.

        Args:
            limit: Maximum number of pages to retrieve
            filter_conditions: Optional filter to apply to the database query
            sorts: Optional sort instructions for the database query

        Returns:
            List of NotionPageManager instances for each page
        """
        await self._ensure_initialized()

        self.logger.debug(
            "Getting up to %d pages with filter: %s, sorts: %s",
            limit,
            filter_conditions,
            sorts,
        )

        pages = await self._query_service.get_pages(
            self.database_id, limit, filter_conditions, sorts
        )

        self.logger.debug(
            "Retrieved %d pages from database %s", len(pages), self.database_id
        )
        return pages

    async def iter_pages(
        self,
        page_size: int = 100,
        filter_conditions: Optional[Dict[str, Any]] = None,
        sorts: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncGenerator[NotionPageManager, None]:
        """
        Asynchronous generator that yields pages from the database.

        Args:
            page_size: Number of pages to fetch per request
            filter_conditions: Optional filter to apply to the database query
            sorts: Optional sort instructions for the database query

        Yields:
            NotionPageManager instances for each page
        """
        await self._ensure_initialized()

        self.logger.debug(
            "Iterating pages with page_size: %d, filter: %s, sorts: %s",
            page_size,
            filter_conditions,
            sorts,
        )

        async for page_manager in self._query_service.iter_pages(
            self.database_id, page_size, filter_conditions, sorts
        ):
            yield page_manager

    async def get_page_manager(self, page_id: str) -> Optional[NotionPageManager]:
        """
        Get a NotionPageManager for a specific page.

        Args:
            page_id: The ID of the page

        Returns:
            NotionPageManager instance or None if the page wasn't found
        """
        await self._ensure_initialized()

        self.logger.debug("Getting page manager for page %s", page_id)

        page_manager = await self._page_service.get_page_manager(page_id)

        if not page_manager:
            self.logger.error("Page %s not found", page_id)

        return page_manager

    async def close(self) -> None:
        """Close the client connection."""
        await self._client.close()



from typing import List, Type
from collections import OrderedDict

from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.registry.block_element_registry import (
    BlockElementRegistry,
)

from notionary.core.converters.elements.paragraph_element import ParagraphElement
from notionary.core.converters.elements.heading_element import HeadingElement
from notionary.core.converters.elements.callout_element import CalloutElement
from notionary.core.converters.elements.code_block_element import CodeBlockElement
from notionary.core.converters.elements.divider_element import DividerElement
from notionary.core.converters.elements.table_element import TableElement
from notionary.core.converters.elements.todo_lists import TodoElement
from notionary.core.converters.elements.list_element import (
    BulletedListElement,
    NumberedListElement,
)
from notionary.core.converters.elements.qoute_element import QuoteElement
from notionary.core.converters.elements.image_element import ImageElement
from notionary.core.converters.elements.video_element import VideoElement
from notionary.core.converters.elements.toggle_element import ToggleElement
from notionary.core.converters.elements.bookmark_element import BookmarkElement
from notionary.core.converters.elements.column_element import ColumnElement


class BlockElementRegistryBuilder:
    """
    True builder for constructing BlockElementRegistry instances.

    This builder allows for incremental construction of registry instances
    with specific configurations of block elements.
    """

    def __init__(self):
        """Initialize a new builder with an empty element list."""
        # Use OrderedDict to maintain insertion order while ensuring uniqueness
        self._elements = OrderedDict()

    # Profile methods - create a base configuration

    @classmethod
    def start_empty(cls) -> "BlockElementRegistryBuilder":
        """
        Start with a completely empty registry builder.

        Returns:
            A new builder instance with no elements
        """
        return cls()

    @classmethod
    def start_minimal(cls) -> "BlockElementRegistryBuilder":
        """
        Start with a minimal set of essential elements.

        Returns:
            A new builder instance with basic elements
        """
        builder = cls()
        return (
            builder.add_element(HeadingElement)
            .add_element(BulletedListElement)
            .add_element(NumberedListElement)
            .add_element(ParagraphElement)
        )  # Add paragraph last as fallback

    @classmethod
    def start_standard(cls) -> "BlockElementRegistryBuilder":
        """
        Start with all standard elements in recommended order.

        Returns:
            A new builder instance with all standard elements
        """
        builder = cls()
        return (
            builder.add_element(HeadingElement)
            .add_element(CalloutElement)
            .add_element(CodeBlockElement)
            .add_element(DividerElement)
            .add_element(TableElement)
            .add_element(ColumnElement)
            .add_element(BulletedListElement)
            .add_element(NumberedListElement)
            .add_element(ToggleElement)
            .add_element(QuoteElement)
            .add_element(TodoElement)
            .add_element(BookmarkElement)
            .add_element(ImageElement)
            .add_element(VideoElement)
            .add_element(ParagraphElement)
        )  # Add paragraph last as fallback

    # Element manipulation methods

    def add_element(
        self, element_class: Type[NotionBlockElement]
    ) -> "BlockElementRegistryBuilder":
        """
        Add an element class to the registry configuration.
        If the element already exists, it's moved to the end.

        Args:
            element_class: The element class to add

        Returns:
            Self for method chaining
        """
        # Remove if exists (to update the order) and add to the end
        self._elements.pop(element_class.__name__, None)
        self._elements[element_class.__name__] = element_class
        return self

    def add_elements(
        self, element_classes: List[Type[NotionBlockElement]]
    ) -> "BlockElementRegistryBuilder":
        """
        Add multiple element classes to the registry configuration.

        Args:
            element_classes: List of element classes to add

        Returns:
            Self for method chaining
        """
        for element_class in element_classes:
            self.add_element(element_class)
        return self

    def remove_element(
        self, element_class: Type[NotionBlockElement]
    ) -> "BlockElementRegistryBuilder":
        """
        Remove an element class from the registry configuration.

        Args:
            element_class: The element class to remove

        Returns:
            Self for method chaining
        """
        self._elements.pop(element_class.__name__, None)
        return self

    def move_element_to_end(
        self, element_class: Type[NotionBlockElement]
    ) -> "BlockElementRegistryBuilder":
        """
        Move an existing element to the end of the registry.
        If the element doesn't exist, it will be added.

        Args:
            element_class: The element class to move

        Returns:
            Self for method chaining
        """
        return self.add_element(element_class)  # add_element already handles this logic

    def ensure_paragraph_at_end(self) -> "BlockElementRegistryBuilder":
        """
        Ensure ParagraphElement is the last element in the registry.
        If it doesn't exist, it will be added.

        Returns:
            Self for method chaining
        """
        return self.move_element_to_end(ParagraphElement)

    # Specialized configuration methods

    def with_list_support(self) -> "BlockElementRegistryBuilder":
        """
        Add support for list elements.

        Returns:
            Self for method chaining
        """
        return self.add_element(BulletedListElement).add_element(NumberedListElement)

    def with_code_support(self) -> "BlockElementRegistryBuilder":
        """
        Add support for code blocks.

        Returns:
            Self for method chaining
        """
        return self.add_element(CodeBlockElement)

    def with_table_support(self) -> "BlockElementRegistryBuilder":
        """
        Add support for tables.

        Returns:
            Self for method chaining
        """
        return self.add_element(TableElement)

    def with_rich_content(self) -> "BlockElementRegistryBuilder":
        """
        Add support for rich content elements (callouts, toggles, etc.).

        Returns:
            Self for method chaining
        """
        return (
            self.add_element(CalloutElement)
            .add_element(ToggleElement)
            .add_element(QuoteElement)
        )

    def with_media_support(self) -> "BlockElementRegistryBuilder":
        """
        Add support for media elements (images, videos).

        Returns:
            Self for method chaining
        """
        return self.add_element(ImageElement).add_element(VideoElement)

    def with_task_support(self) -> "BlockElementRegistryBuilder":
        """
        Add support for task-related elements (todos).

        Returns:
            Self for method chaining
        """
        return self.add_element(TodoElement)

    def build(self) -> BlockElementRegistry:
        """
        Build and return the configured BlockElementRegistry instance.

        Returns:
            A configured BlockElementRegistry instance
        """
        registry = BlockElementRegistry()

        # Add elements in the recorded order
        for element_class in self._elements.values():
            registry.register(element_class)

        return registry

    @classmethod
    def create_standard_registry(cls) -> BlockElementRegistry:
        """
        Factory method to directly create a standard registry.

        Returns:
            A fully configured registry instance
        """
        return cls.start_standard().build()

    @classmethod
    def create_minimal_registry(cls) -> BlockElementRegistry:
        """
        Factory method to directly create a minimal registry.

        Returns:
            A minimal registry instance
        """
        return cls.start_minimal().build()

    @classmethod
    def create_custom_registry(
        cls, element_classes: List[Type[NotionBlockElement]]
    ) -> BlockElementRegistry:
        """
        Factory method to directly create a custom registry.

        Args:
            element_classes: List of element classes to register

        Returns:
            A custom configured registry instance
        """
        return cls().add_elements(element_classes).build()



# File: elements/dividers.py

from typing import Dict, Any, Optional
from typing_extensions import override
import re

from notionary.core.converters.elements.notion_block_element import NotionBlockElement


class DividerElement(NotionBlockElement):
    """
    Handles conversion between Markdown horizontal dividers and Notion divider blocks.

    Markdown divider syntax:
    - Three or more hyphens (---) on a line by themselves
    """

    PATTERN = re.compile(r"^\s*-{3,}\s*$")

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text is a markdown divider."""
        return bool(DividerElement.PATTERN.match(text))

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion divider."""
        return block.get("type") == "divider"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown divider to Notion divider block."""
        if not DividerElement.match_markdown(text):
            return None

        return {"type": "divider", "divider": {}}

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion divider block to markdown divider."""
        if block.get("type") != "divider":
            return None

        return "---"

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """
        Returns a dictionary with all information needed for LLM prompts about this element.
        Includes description, usage guidance, syntax options, and examples.
        """
        return {
            "description": "Creates a horizontal divider line that visually separates sections of content.",
            "when_to_use": "Use dividers when you want to create clear visual breaks between different sections or topics in your document. Dividers help improve readability by organizing content into distinct sections without requiring headings.",
            "syntax": ["---"],
            "notes": [
                "Dividers must be on their own line with no other content",
                "Dividers work well when combined with headings to clearly separate major document sections",
            ],
            "examples": [
                "## Introduction\nThis is the introduction section of the document.\n\n---\n\n## Main Content\nThis is the main content section.",
                "Task List:\n- Complete project proposal\n- Review feedback\n\n---\n\nMeeting Notes:\n- Discussed timeline\n- Assigned responsibilities",
            ],
        }



# Import converters
from .markdown_to_notion_converter import MarkdownToNotionConverter
from .notion_to_markdown_converter import NotionToMarkdownConverter

# Import registry classes
from .registry.block_element_registry import BlockElementRegistry
from .registry.block_element_registry_builder import BlockElementRegistryBuilder

# Import elements for type hints and direct use
from .elements.paragraph_element import ParagraphElement
from .elements.heading_element import HeadingElement
from .elements.callout_element import CalloutElement
from .elements.code_block_element import CodeBlockElement
from .elements.divider_element import DividerElement
from .elements.table_element import TableElement
from .elements.todo_lists import TodoElement
from .elements.list_element import BulletedListElement, NumberedListElement
from .elements.qoute_element import QuoteElement
from .elements.image_element import ImageElement
from .elements.video_element import VideoElement
from .elements.toggle_element import ToggleElement
from .elements.bookmark_element import BookmarkElement
from .elements.column_element import ColumnElement

default_registry = BlockElementRegistryBuilder.create_standard_registry()

# Define what to export
__all__ = [
    "BlockElementRegistry",
    "BlockElementRegistryBuilder",
    "MarkdownToNotionConverter",
    "NotionToMarkdownConverter",
    "default_registry",
    # Element classes
    "ParagraphElement",
    "HeadingElement",
    "CalloutElement",
    "CodeBlockElement",
    "DividerElement",
    "TableElement",
    "TodoElement",
    "QuoteElement",
    "BulletedListElement",
    "NumberedListElement",
    "ImageElement",
    "VideoElement",
    "ToggleElement",
    "BookmarkElement",
    "ColumnElement",
]



# File: elements/tables.py

from typing import Dict, Any, Optional, List, Tuple
from typing_extensions import override
import re
from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter


class TableElement(NotionBlockElement):
    """
    Handles conversion between Markdown tables and Notion table blocks.

    Markdown table syntax:
    | Header 1 | Header 2 | Header 3 |
    | -------- | -------- | -------- |
    | Cell 1   | Cell 2   | Cell 3   |
    | Cell 4   | Cell 5   | Cell 6   |

    The second line with dashes and optional colons defines column alignment.
    """

    # Patterns for detecting Markdown tables
    ROW_PATTERN = re.compile(r"^\s*\|(.+)\|\s*$")
    SEPARATOR_PATTERN = re.compile(r"^\s*\|([\s\-:|]+)\|\s*$")

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text contains a markdown table."""
        lines = text.split("\n")

        if len(lines) < 3:
            return False

        for i, line in enumerate(lines[:-2]):
            if (
                TableElement.ROW_PATTERN.match(line)
                and TableElement.SEPARATOR_PATTERN.match(lines[i + 1])
                and TableElement.ROW_PATTERN.match(lines[i + 2])
            ):
                return True

        return False

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion table."""
        return block.get("type") == "table"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown table to Notion table block."""
        if not TableElement.match_markdown(text):
            return None

        lines = text.split("\n")

        table_start = TableElement._find_table_start(lines)
        if table_start is None:
            return None

        table_end = TableElement._find_table_end(lines, table_start)
        table_lines = lines[table_start:table_end]

        rows = TableElement._extract_table_rows(table_lines)
        if not rows:
            return None

        column_count = len(rows[0])
        TableElement._normalize_row_lengths(rows, column_count)

        return {
            "type": "table",
            "table": {
                "table_width": column_count,
                "has_column_header": True,
                "has_row_header": False,
                "children": TableElement._create_table_rows(rows),
            },
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion table block to markdown table."""
        if block.get("type") != "table":
            return None

        table_data = block.get("table", {})
        children = block.get("children", [])

        if not children:
            table_width = table_data.get("table_width", 3)

            header = (
                "| " + " | ".join([f"Column {i+1}" for i in range(table_width)]) + " |"
            )
            separator = (
                "| " + " | ".join(["--------" for _ in range(table_width)]) + " |"
            )
            data_row = (
                "| " + " | ".join(["        " for _ in range(table_width)]) + " |"
            )

            table_rows = [header, separator, data_row]
            return "\n".join(table_rows)

        table_rows = []
        header_processed = False

        for child in children:
            if child.get("type") != "table_row":
                continue

            row_data = child.get("table_row", {})
            cells = row_data.get("cells", [])

            row_cells = []
            for cell in cells:
                cell_text = TextInlineFormatter.extract_text_with_formatting(cell)
                row_cells.append(cell_text or "")

            row = "| " + " | ".join(row_cells) + " |"
            table_rows.append(row)

            if not header_processed and table_data.get("has_column_header", True):
                header_processed = True
                separator = (
                    "| " + " | ".join(["--------" for _ in range(len(cells))]) + " |"
                )
                table_rows.append(separator)

        if not table_rows:
            return None

        if len(table_rows) == 1 and table_data.get("has_column_header", True):
            cells_count = len(children[0].get("table_row", {}).get("cells", []))
            separator = (
                "| " + " | ".join(["--------" for _ in range(cells_count)]) + " |"
            )
            table_rows.insert(1, separator)

        return "\n".join(table_rows)

    @override
    @staticmethod
    def is_multiline() -> bool:
        """Indicates if this element handles content that spans multiple lines."""
        return True

    @staticmethod
    def _find_table_start(lines: List[str]) -> Optional[int]:
        """Find the start index of a table in the lines."""
        for i in range(len(lines) - 2):
            if (
                TableElement.ROW_PATTERN.match(lines[i])
                and TableElement.SEPARATOR_PATTERN.match(lines[i + 1])
                and TableElement.ROW_PATTERN.match(lines[i + 2])
            ):
                return i
        return None

    @staticmethod
    def _find_table_end(lines: List[str], start_idx: int) -> int:
        """Find the end index of a table, starting from start_idx."""
        end_idx = start_idx + 3  # Minimum: Header, Separator, one data row
        while end_idx < len(lines) and TableElement.ROW_PATTERN.match(lines[end_idx]):
            end_idx += 1
        return end_idx

    @staticmethod
    def _extract_table_rows(table_lines: List[str]) -> List[List[str]]:
        """Extract row contents from table lines, excluding separator line."""
        rows = []
        for i, line in enumerate(table_lines):
            if i != 1 and TableElement.ROW_PATTERN.match(line):  # Skip separator line
                cells = TableElement._parse_table_row(line)
                if cells:
                    rows.append(cells)
        return rows

    @staticmethod
    def _normalize_row_lengths(rows: List[List[str]], column_count: int) -> None:
        """Normalize row lengths to the specified column count."""
        for row in rows:
            if len(row) < column_count:
                row.extend([""] * (column_count - len(row)))
            elif len(row) > column_count:
                del row[column_count:]

    @staticmethod
    def _parse_table_row(row_text: str) -> List[str]:
        """Convert table row text to cell contents."""
        row_content = row_text.strip()

        if row_content.startswith("|"):
            row_content = row_content[1:]
        if row_content.endswith("|"):
            row_content = row_content[:-1]

        return [cell.strip() for cell in row_content.split("|")]

    @staticmethod
    def _create_table_rows(rows: List[List[str]]) -> List[Dict[str, Any]]:
        """Create Notion table rows from cell contents."""
        table_rows = []

        for row in rows:
            cells_data = []

            for cell_content in row:
                rich_text = TextInlineFormatter.parse_inline_formatting(cell_content)

                if not rich_text:
                    rich_text = [
                        {
                            "type": "text",
                            "text": {"content": ""},
                            "annotations": {
                                "bold": False,
                                "italic": False,
                                "strikethrough": False,
                                "underline": False,
                                "code": False,
                                "color": "default",
                            },
                            "plain_text": "",
                            "href": None,
                        }
                    ]

                cells_data.append(rich_text)

            table_rows.append({"type": "table_row", "table_row": {"cells": cells_data}})

        return table_rows

    @staticmethod
    def find_matches(text: str) -> List[Tuple[int, int, Dict[str, Any]]]:
        """
        Find all tables in the text and return their positions.

        Args:
            text: The text to search in

        Returns:
            List of tuples with (start_pos, end_pos, block)
        """
        matches = []
        lines = text.split("\n")

        i = 0
        while i < len(lines) - 2:
            if (
                TableElement.ROW_PATTERN.match(lines[i])
                and TableElement.SEPARATOR_PATTERN.match(lines[i + 1])
                and TableElement.ROW_PATTERN.match(lines[i + 2])
            ):

                start_line = i
                end_line = TableElement._find_table_end(lines, start_line)

                start_pos = TableElement._calculate_position(lines, 0, start_line)
                end_pos = start_pos + TableElement._calculate_position(
                    lines, start_line, end_line
                )

                table_text = "\n".join(lines[start_line:end_line])
                table_block = TableElement.markdown_to_notion(table_text)

                if table_block:
                    matches.append((start_pos, end_pos, table_block))

                i = end_line
            else:
                i += 1

        return matches

    @staticmethod
    def _calculate_position(lines: List[str], start: int, end: int) -> int:
        """Calculate the text position in characters from line start to end."""
        position = 0
        for i in range(start, end):
            position += len(lines[i]) + 1  # +1 for newline
        return position

    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """Returns information for LLM prompts about this element."""
        return {
            "description": "Creates formatted tables with rows and columns for structured data.",
            "when_to_use": "Use tables to organize and present structured data in a grid format, making information easier to compare and analyze. Tables are ideal for data sets, comparison charts, pricing information, or any content that benefits from columnar organization.",
            "notes": [
                "The header row is required and will be displayed differently in Notion",
                "The separator row with dashes is required to define the table structure",
                "Table cells support inline formatting such as **bold** and *italic*",
            ],
            "examples": [
                "| Product | Price | Stock |\n| ------- | ----- | ----- |\n| Widget A | $10.99 | 42 |\n| Widget B | $14.99 | 27 |",
                "| Name | Role | Department |\n| ---- | ---- | ---------- |\n| John Smith | Manager | Marketing |\n| Jane Doe | Director | Sales |",
            ],
        }



from typing import Dict, Any, Optional
from typing_extensions import override
import re

from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter


class HeadingElement(NotionBlockElement):
    """Handles conversion between Markdown headings and Notion heading blocks."""

    PATTERN = re.compile(r"^(#{1,6})\s(.+)$")

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text is a markdown heading."""
        return bool(HeadingElement.PATTERN.match(text))

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion heading."""
        block_type: str = block.get("type", "")
        return block_type.startswith("heading_") and block_type[-1] in "123456"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown heading to Notion heading block."""
        header_match = HeadingElement.PATTERN.match(text)
        if not header_match:
            return None

        level = len(header_match.group(1))
        content = header_match.group(2)

        # Import here to avoid circular imports

        return {
            "type": f"heading_{level}",
            f"heading_{level}": {
                "rich_text": TextInlineFormatter.parse_inline_formatting(content)
            },
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion heading block to markdown heading."""
        block_type = block.get("type", "")

        if not block_type.startswith("heading_"):
            return None

        try:
            level = int(block_type[-1])
            if not 1 <= level <= 6:
                return None
        except ValueError:
            return None

        heading_data = block.get(block_type, {})
        rich_text = heading_data.get("rich_text", [])

        text = TextInlineFormatter.extract_text_with_formatting(rich_text)
        prefix = "#" * level
        return f"{prefix} {text or ''}"

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @override
    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """
        Returns a dictionary with all information needed for LLM prompts about this element.
        """
        return {
            "description": "Use Markdown headings (#, ##, ###, etc.) to structure content hierarchically.",
            "when_to_use": "Use to group content into sections and define a visual hierarchy.",
        }



from typing import TypeVar
from functools import wraps

T = TypeVar("T")


def singleton(cls: T) -> T:
    """
    Decorator zur Implementierung des Singleton-Musters.
    Stellt sicher, dass nur eine Instanz der Klasse existiert.
    """
    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance



from typing import Dict, Any, Optional, List, Type

from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter


class BlockElementRegistry:
    """Registry of elements that can convert between Markdown and Notion."""

    def __init__(self, elements=None):
        """
        Initialize a new registry instance.

        Args:
            elements: Optional list of NotionBlockElement classes to register at creation
        """
        self._elements = []

        # Register initial elements if provided
        if elements:
            for element in elements:
                self.register(element)

    def register(self, element_class: Type[NotionBlockElement]):
        """Register an element class."""
        self._elements.append(element_class)
        return self

    def deregister(self, element_class: Type[NotionBlockElement]) -> bool:
        """
        Deregister an element class.

        Args:
            element_class: The element class to remove from the registry

        Returns:
            bool: True if the element was removed, False if it wasn't in the registry
        """
        if element_class in self._elements:
            self._elements.remove(element_class)
            return True
        return False

    def clear(self):
        """Clear the registry completely."""
        self._elements.clear()
        return self

    def find_markdown_handler(self, text: str) -> Optional[Type[NotionBlockElement]]:
        """Find an element that can handle the given markdown text."""
        for element in self._elements:
            if element.match_markdown(text):
                return element
        return None

    def find_notion_handler(
        self, block: Dict[str, Any]
    ) -> Optional[Type[NotionBlockElement]]:
        """Find an element that can handle the given Notion block."""
        for element in self._elements:
            if element.match_notion(block):
                return element
        return None

    def markdown_to_notion(self, text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown to Notion block using registered elements."""
        handler = self.find_markdown_handler(text)
        if handler:
            return handler.markdown_to_notion(text)
        return None

    def notion_to_markdown(self, block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion block to markdown using registered elements."""
        handler = self.find_notion_handler(block)
        if handler:
            return handler.notion_to_markdown(block)
        return None

    def get_multiline_elements(self) -> List[Type[NotionBlockElement]]:
        """Get all registered multiline elements."""
        return [element for element in self._elements if element.is_multiline()]

    def get_elements(self) -> List[Type[NotionBlockElement]]:
        """Get all registered elements."""
        return self._elements.copy()

    def generate_llm_prompt(self) -> str:
        """
        Generates an LLM system prompt that describes the Markdown syntax of all registered elements.

        TextInlineFormatter is automatically added if not already registered.

        Returns:
            A complete system prompt for an LLM that should understand Notion-Markdown syntax
        """
        # Create a copy of registered elements
        element_classes = self._elements.copy()
        print("Elements in registry:", element_classes)

        formatter_names = [e.__name__ for e in element_classes]
        if "TextInlineFormatter" not in formatter_names:
            element_classes = [TextInlineFormatter] + element_classes

        return MarkdownSyntaxPromptBuilder.generate_system_prompt(element_classes)


class MarkdownSyntaxPromptBuilder:
    """
    Generator for LLM system prompts that describe Notion-Markdown syntax.

    This class extracts information about supported Markdown patterns
    and formats them optimally for LLMs.
    """

    # Standard system prompt template
    SYSTEM_PROMPT_TEMPLATE = """You are a knowledgeable assistant that helps users create content for Notion pages.
Notion supports standard Markdown with some special extensions for creating rich content.

{element_docs}

Important usage guidelines:

1. The backtick code fence syntax (```) should ONLY be used when creating actual code blocks or diagrams.
Do not wrap examples or regular content in backticks unless you're showing code.

2. Use inline formatting (bold, italic, highlights, etc.) across all content to enhance readability.
The highlight syntax (==text== and ==color:text==) is especially useful for emphasizing important points.

3. Notion's extensions to Markdown (like callouts, bookmarks, toggles) provide richer formatting options
than standard Markdown while maintaining the familiar Markdown syntax for basic elements.

4. You can use these Markdown extensions alongside standard Markdown to create visually appealing
and well-structured content.

5. Remember that features like highlighting with ==yellow:important== work in all text blocks including
paragraphs, lists, quotes, etc.
"""

    @staticmethod
    def generate_element_doc(element_class: Type[NotionBlockElement]) -> str:
        """
        Generates documentation for a specific NotionBlockElement.

        Uses the element's get_llm_prompt_content method if available.
        """
        class_name = element_class.__name__
        element_name = class_name.replace("Element", "")

        # Start with element name as header
        result = [f"## {element_name}"]

        # Use get_llm_prompt_content if available
        if hasattr(element_class, "get_llm_prompt_content") and callable(
            getattr(element_class, "get_llm_prompt_content")
        ):
            content = element_class.get_llm_prompt_content()

            if content.get("description"):
                result.append(content["description"])

            if content.get("syntax"):
                result.append("\n### Syntax:")
                for syntax_item in content["syntax"]:
                    result.append(f"{syntax_item}")

            if content.get("examples"):
                result.append("\n### Examples:")
                for example in content["examples"]:
                    result.append(example)

            # Add any additional custom sections
            for key, value in content.items():
                if key not in ["description", "syntax", "examples"] and isinstance(
                    value, str
                ):
                    result.append(f"\n### {key.replace('_', ' ').title()}:")
                    result.append(value)

        return "\n".join(result)

    @classmethod
    def generate_element_docs(
        cls,
        element_classes: List[Type[NotionBlockElement]],
    ) -> str:
        """
        Generates complete documentation for all provided element classes.

        Args:
            element_classes: List of NotionBlockElement classes

        Returns:
            Documentation text for all elements
        """
        docs = [
            "# Custom Markdown Syntax for Notion Blocks",
            "The following custom Markdown patterns are supported for creating Notion blocks:",
        ]

        text_formatter = None
        other_elements = []

        for element in element_classes:
            if element.__name__ == "TextInlineFormatter":
                text_formatter = element
            else:
                other_elements.append(element)

        if text_formatter:
            docs.append("\n" + cls.generate_element_doc(text_formatter))

        for element in other_elements:
            if element.__name__ != "InlineFormattingElement":
                docs.append("\n" + cls.generate_element_doc(element))

        return "\n".join(docs)

    @classmethod
    def generate_system_prompt(
        cls,
        element_classes: List[Type[NotionBlockElement]],
    ) -> str:
        """
        Generates a complete system prompt for LLMs.

        Args:
            element_classes: List of element classes to document

        Returns:
            Complete system prompt for an LLM
        """
        element_docs = cls.generate_element_docs(element_classes)

        return cls.SYSTEM_PROMPT_TEMPLATE.format(element_docs=element_docs)



import re
from typing import Dict, Any, Optional, List
from typing_extensions import override
from notionary.core.converters.elements.notion_block_element import NotionBlockElement


class ImageElement(NotionBlockElement):
    """
    Handles conversion between Markdown images and Notion image blocks.

    Markdown image syntax:
    - ![Caption](https://example.com/image.jpg) - Basic image with caption
    - ![](https://example.com/image.jpg) - Image without caption
    - ![Caption](https://example.com/image.jpg "alt text") - Image with caption and alt text
    """

    # Regex pattern for image syntax with optional alt text
    PATTERN = re.compile(
        r"^\!\[(.*?)\]"  # ![Caption] part
        + r'\((https?://[^\s"]+)'  # (URL part
        + r'(?:\s+"([^"]+)")?'  # Optional alt text in quotes
        + r"\)$"  # closing parenthesis
    )

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text is a markdown image."""
        return text.strip().startswith("![") and bool(
            ImageElement.PATTERN.match(text.strip())
        )

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion image."""
        return block.get("type") == "image"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown image to Notion image block."""
        image_match = ImageElement.PATTERN.match(text.strip())
        if not image_match:
            return None

        caption = image_match.group(1)
        url = image_match.group(2)

        if not url:
            return None

        # Prepare the image block
        image_block = {
            "type": "image",
            "image": {"type": "external", "external": {"url": url}},
        }

        # Add caption if provided
        if caption:
            image_block["image"]["caption"] = [
                {"type": "text", "text": {"content": caption}}
            ]

        return image_block

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion image block to markdown image."""
        if block.get("type") != "image":
            return None

        image_data = block.get("image", {})

        # Handle both external and file (uploaded) images
        if image_data.get("type") == "external":
            url = image_data.get("external", {}).get("url", "")
        elif image_data.get("type") == "file":
            url = image_data.get("file", {}).get("url", "")
        else:
            return None

        if not url:
            return None

        # Extract caption if available
        caption = ""
        caption_rich_text = image_data.get("caption", [])
        if caption_rich_text:
            caption = ImageElement._extract_text_content(caption_rich_text)

        return f"![{caption}]({url})"

    @staticmethod
    def _extract_text_content(rich_text: List[Dict[str, Any]]) -> str:
        """Extract plain text content from Notion rich_text elements."""
        result = ""
        for text_obj in rich_text:
            if text_obj.get("type") == "text":
                result += text_obj.get("text", {}).get("content", "")
            elif "plain_text" in text_obj:
                result += text_obj.get("plain_text", "")
        return result

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """
        Returns a dictionary with all information needed for LLM prompts about this element.
        Includes description, usage guidance, syntax options, and examples.
        """
        return {
            "description": "Embeds an image from an external URL into your document.",
            "when_to_use": "Use images to include visual content such as diagrams, screenshots, charts, photos, or illustrations that enhance your document. Images can make complex information easier to understand, create visual interest, or provide evidence for your points.",
            "syntax": [
                "![](https://example.com/image.jpg) - Image without caption",
                "![Caption text](https://example.com/image.jpg) - Image with caption",
                '![Caption text](https://example.com/image.jpg "Alt text") - Image with caption and alt text',
            ],
            "examples": [
                "![Data visualization showing monthly trends](https://example.com/chart.png)",
                "![](https://example.com/screenshot.jpg)",
                '![Company logo](https://company.com/logo.png "Company Inc. logo")',
            ],
        }



from typing import Dict, Any, Optional
from typing_extensions import override

from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter


class ParagraphElement(NotionBlockElement):
    """Handles conversion between Markdown paragraphs and Notion paragraph blocks."""

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """
        Check if text is a markdown paragraph.
        Paragraphs are essentially any text that isn't matched by other block elements.
        Since paragraphs are the fallback element, this always returns True.
        """
        return True

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion paragraph."""
        return block.get("type") == "paragraph"

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown paragraph to Notion paragraph block."""
        if not text.strip():
            return None

        return {
            "type": "paragraph",
            "paragraph": {
                "rich_text": TextInlineFormatter.parse_inline_formatting(text)
            },
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion paragraph block to markdown paragraph."""
        if block.get("type") != "paragraph":
            return None

        paragraph_data = block.get("paragraph", {})
        rich_text = paragraph_data.get("rich_text", [])

        text = TextInlineFormatter.extract_text_with_formatting(rich_text)
        return text if text else None

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """Returns information for LLM prompts about this element."""
        return {
            "description": "Creates standard paragraph blocks for regular text content.",
            "when_to_use": "Use paragraphs for normal text content. Paragraphs are the default block type and will be used when no other specific formatting is applied.",
            "syntax": ["Just write text normally without any special prefix"],
            "notes": [
                "Paragraphs support inline formatting like **bold**, *italic*, ~~strikethrough~~, `code`, and [links](url)"
            ],
            "examples": [
                "This is a simple paragraph with plain text.",
                "This paragraph has **bold** and *italic* formatting.",
            ],
        }



from typing import (
    AsyncGenerator,
    Dict,
    List,
    Optional,
    Any,
    TypedDict,
    Union,
    cast,
    Literal,
)
from notionary.core.notion_client import NotionClient
from notionary.core.page.notion_page_manager import NotionPageManager
from notionary.util.logging_mixin import LoggingMixin


class NotionTextContent(TypedDict):
    plain_text: str


class NotionTitleProperty(TypedDict):
    type: Literal["title"]
    title: List[NotionTextContent]


class NotionSelectOption(TypedDict):
    name: str
    id: Optional[str]
    color: Optional[str]


class NotionSelectProperty(TypedDict):
    type: Literal["select"]
    select: Dict[str, List[NotionSelectOption]]


class NotionMultiSelectProperty(TypedDict):
    type: Literal["multi_select"]
    multi_select: Dict[str, List[NotionSelectOption]]


class NotionStatusProperty(TypedDict):
    type: Literal["status"]
    status: Dict[str, List[NotionSelectOption]]


class NotionRelationProperty(TypedDict):
    type: Literal["relation"]
    relation: Dict[str, str]


class NotionNumberProperty(TypedDict):
    type: Literal["number"]
    number: Dict[str, Any]


NotionPropertyType = Union[
    NotionTitleProperty,
    NotionSelectProperty,
    NotionMultiSelectProperty,
    NotionStatusProperty,
    NotionRelationProperty,
    NotionNumberProperty,
    Dict[str, Any],  # Fallback
]


class RelationOption(TypedDict):
    id: str
    title: str


class NotionDatabaseAccessor(LoggingMixin):
    """
    A utility class that provides methods to access Notion databases.
    Focused on efficient, paginated access to databases without unnecessary complexity.
    """

    def __init__(self, client: Optional[NotionClient] = None) -> None:
        """
        Initialize the accessor with a NotionClient.

        Args:
            client: NotionClient instance for API communication
        """
        self._client = client if client else NotionClient()
        self.logger.info("NotionDatabaseAccessor initialized")

    async def iter_databases(
        self, page_size: int = 100
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Asynchronous generator that yields Notion databases one by one.

        Uses the Notion API to provide paginated access to all databases
        without loading all of them into memory at once.

        Args:
            page_size: The number of databases to fetch per request

        Yields:
            Individual database objects from the Notion API
        """
        start_cursor: Optional[str] = None

        while True:
            body: Dict[str, Any] = {
                "filter": {"value": "database", "property": "object"},
                "page_size": page_size,
            }

            if start_cursor:
                body["start_cursor"] = start_cursor

            result = await self._client.post("search", data=body)

            if not result or "results" not in result:
                self.logger.error("Error fetching databases")
                break

            for database in result["results"]:
                yield database

            if "has_more" in result and result["has_more"] and "next_cursor" in result:
                start_cursor = result["next_cursor"]
            else:
                break

    async def get_database(self, database_id: str) -> Optional[Dict[str, Any]]:
        """
        Get the details for a specific database.

        Args:
            database_id: The ID of the database

        Returns:
            Database details or None if not found
        """
        db_details = await self._client.get(f"databases/{database_id}")
        if not db_details:
            self.logger.error("Failed to retrieve database %s", database_id)
            return None

        return db_details

    def extract_database_title(self, database: Dict[str, Any]) -> str:
        """
        Extract the database title from a Notion API response.

        Args:
            database: The database object from the Notion API

        Returns:
            The extracted title or "Untitled" if no title is found
        """
        title = "Untitled"

        if "title" in database:
            title_parts = []
            for text_obj in database["title"]:
                if "plain_text" in text_obj:
                    title_parts.append(text_obj["plain_text"])

            if title_parts:
                title = "".join(title_parts)

        return title


class NotionDatabaseSchema:
    """
    Represents the schema of a specific Notion database.
    Manages property information, options, and relations for a single database.
    """

    def __init__(self, database_id: str, client: NotionClient) -> None:
        """
        Initialize a database schema handler for a specific database.

        Args:
            database_id: The ID of the database
            client: An instance of NotionClient for API requests
        """
        self.database_id: str = database_id
        self._client: NotionClient = client
        self._properties: Dict[str, NotionPropertyType] = {}
        self._loaded: bool = False

    async def load(self) -> bool:
        """
        Load the database schema from the Notion API.

        Returns:
            True if the schema was loaded successfully, False otherwise
        """
        if self._loaded:
            return True

        db_details = await self._client.get(f"databases/{self.database_id}")
        if not db_details or "properties" not in db_details:
            return False

        self._properties = db_details["properties"]
        self._loaded = True
        return True

    async def get_property_types(self) -> Dict[str, str]:
        """
        Get a mapping of property names to their types.

        Returns:
            A dictionary mapping property names to types
        """
        if not self._loaded:
            await self.load()

        return {name: prop.get("type", "") for name, prop in self._properties.items()}

    async def get_select_options(self, property_name: str) -> List[NotionSelectOption]:
        """
        Get the options for a select, multi_select, or status property.

        Args:
            property_name: The name of the property

        Returns:
            A list of option objects
        """
        if not self._loaded:
            await self.load()

        if property_name not in self._properties:
            return []

        prop = self._properties[property_name]
        prop_type = prop.get("type", "")

        if prop_type not in ["select", "multi_select", "status"]:
            return []

        if prop_type in prop and "options" in prop[prop_type]:
            return cast(List[NotionSelectOption], prop[prop_type]["options"])

        return []

    async def get_relation_options(
        self, property_name: str, limit: int = 100
    ) -> List[RelationOption]:
        """
        Get available options for a relation property (pages in the related database).

        Args:
            property_name: The name of the relation property
            limit: Maximum number of options to retrieve

        Returns:
            List of options with id and title
        """
        related_db_id = await self.get_relation_database_id(property_name)
        if not related_db_id:
            return []

        pages = await self._query_database_pages(related_db_id, limit)
        return self._extract_page_titles_and_ids(pages)

    async def get_relation_database_id(self, property_name: str) -> Optional[str]:
        """
        Get the ID of the related database for a relation property.

        Args:
            property_name: The name of the property

        Returns:
            The ID of the related database or None
        """
        if not self._loaded:
            await self.load()

        if property_name not in self._properties:
            return None

        prop = self._properties[property_name]
        prop_type = prop.get("type", "")

        if prop_type != "relation" or "relation" not in prop:
            return None

        relation_prop = cast(NotionRelationProperty, prop)
        return relation_prop["relation"].get("database_id")

    def _extract_page_titles_and_ids(
        self, pages: List[NotionPageManager]
    ) -> List[RelationOption]:
        """
        Extract titles and IDs from page objects.

        Args:
            pages: List of page objects from the Notion API

        Returns:
            List of dictionaries with id and title for each page
        """
        options: List[RelationOption] = []

        for page_manager in pages:
            page_title = page_manager.title or "Untitled"
            options.append({"id": page_manager.page_id, "title": page_title})

        return options

    async def _query_database_pages(
        self, database_id: str, limit: int = 100
    ) -> List[Dict[str, Any]]:
        """
        Returns:
            List of page objects from the Notion API
        """
        pages: List[Dict[str, Any]] = []
        count = 0

        async for page in self.iter_database_pages(
            database_id=database_id, page_size=min(limit, 100)
        ):
            pages.append(page)
            count += 1

            if count >= limit:
                break

        return pages

    async def iter_database_pages(
        self,
        database_id: Optional[str] = None,
        page_size: int = 100,
        filter_conditions: Optional[Dict[str, Any]] = None,
        sorts: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncGenerator[NotionPageManager, None]:
        """
        Asynchronous generator that yields pages from a Notion database one by one.

        Uses the Notion API to provide paginated access to all pages in a database
        without loading all of them into memory at once.

        Args:
            database_id: The ID of the database to query (uses self.database_id if None)
            page_size: The number of pages to fetch per request
            filter_conditions: Optional filter to apply to the database query
            sorts: Optional sort instructions for the database query

        Yields:
            Individual page objects from the Notion API
        """
        db_id = database_id or self.database_id
        if not db_id:
            raise ValueError("No database ID provided")

        start_cursor: Optional[str] = None
        has_more = True

        body: Dict[str, Any] = {"page_size": page_size}

        if filter_conditions:
            body["filter"] = filter_conditions

        if sorts:
            body["sorts"] = sorts

        while has_more:
            current_body = body.copy()
            if start_cursor:
                current_body["start_cursor"] = start_cursor

            result = await self._client.post(
                f"databases/{db_id}/query", data=current_body
            )

            if not result or "results" not in result:
                return

            for page in result["results"]:
                page_id = page.get("id", "")
                title = self._extract_page_title(page)

                notion_page_manager = NotionPageManager(page_id=page_id, title=title)
                yield notion_page_manager

            has_more = result.get("has_more", False)
            start_cursor = result.get("next_cursor") if has_more else None

    def _extract_page_title(self, page: Dict[str, Any]) -> str:
        """
        Extracts the title from a Notion page object.

        Args:
            page: The Notion page object

        Returns:
            The extracted title as a string, or an empty string if no title found
        """
        properties = page.get("properties", {})
        if not properties:
            return ""

        for prop_value in properties.values():
            if prop_value.get("type") != "title":
                continue

            title_array = prop_value.get("title", [])
            if not title_array:
                continue

            return title_array[0].get("plain_text", "")

        return ""



import re
from typing import Dict, Any, Optional
from typing_extensions import override
from notionary.core.converters.elements.notion_block_element import NotionBlockElement
from notionary.core.converters.elements.text_inline_formatter import TextInlineFormatter


class BulletedListElement(NotionBlockElement):
    """Class for converting between Markdown bullet lists and Notion bulleted list items."""

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown bulleted list item to Notion block."""
        pattern = re.compile(
            r"^(\s*)[*\-+]\s+(?!\[[ x]\])(.+)$"
        )  # Avoid matching todo items
        list_match = pattern.match(text)
        if not list_match:
            return None

        content = list_match.group(2)

        # Use parse_inline_formatting to handle rich text
        rich_text = TextInlineFormatter.parse_inline_formatting(content)

        return {
            "type": "bulleted_list_item",
            "bulleted_list_item": {"rich_text": rich_text, "color": "default"},
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion bulleted list item block to markdown."""
        if block.get("type") != "bulleted_list_item":
            return None

        rich_text = block.get("bulleted_list_item", {}).get("rich_text", [])
        content = TextInlineFormatter.extract_text_with_formatting(rich_text)

        return f"- {content}"

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if this element can handle the given markdown text."""
        pattern = re.compile(r"^(\s*)[*\-+]\s+(?!\[[ x]\])(.+)$")
        return bool(pattern.match(text))

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if this element can handle the given Notion block."""
        return block.get("type") == "bulleted_list_item"

    @override
    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """Returns information for LLM prompts about this element."""
        return {
            "description": "Creates bulleted list items for unordered lists.",
            "when_to_use": "Use for lists where order doesn't matter, such as features, options, or items without hierarchy.",
            "syntax": ["- Item text", "* Item text", "+ Item text"],
            "examples": ["- First item\n- Second item\n- Third item"],
        }


class NumberedListElement:
    """Class for converting between Markdown numbered lists and Notion numbered list items."""

    @override
    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown numbered list item to Notion block."""
        pattern = re.compile(r"^\s*(\d+)\.\s+(.+)$")
        numbered_match = pattern.match(text)
        if not numbered_match:
            return None

        content = numbered_match.group(2)

        # Use parse_inline_formatting to handle rich text
        rich_text = TextInlineFormatter.parse_inline_formatting(content)

        return {
            "type": "numbered_list_item",
            "numbered_list_item": {"rich_text": rich_text, "color": "default"},
        }

    @override
    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion numbered list item block to markdown."""
        if block.get("type") != "numbered_list_item":
            return None

        rich_text = block.get("numbered_list_item", {}).get("rich_text", [])
        content = TextInlineFormatter.extract_text_with_formatting(rich_text)

        return f"1. {content}"

    @override
    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if this element can handle the given markdown text."""
        pattern = re.compile(r"^\s*\d+\.\s+(.+)$")
        return bool(pattern.match(text))

    @override
    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if this element can handle the given Notion block."""
        return block.get("type") == "numbered_list_item"

    @override
    @staticmethod
    def is_multiline() -> bool:
        return False

    @override
    @classmethod
    def get_llm_prompt_content(cls) -> dict:
        """Returns information for LLM prompts about this element."""
        return {
            "description": "Creates numbered list items for ordered sequences.",
            "when_to_use": "Use for lists where order matters, such as steps, rankings, or sequential items.",
            "syntax": ["1. Item text"],
            "examples": ["1. First step\n2. Second step\n3. Third step"],
        }



from typing import Any, Dict, List, Optional

from notionary.core.converters.markdown_to_notion_converter import (
    MarkdownToNotionConverter,
)
from notionary.core.converters.notion_to_markdown_converter import (
    NotionToMarkdownConverter,
)
from notionary.core.converters.registry.block_element_registry import (
    BlockElementRegistry,
)
from notionary.core.notion_client import NotionClient
from notionary.util.logging_mixin import LoggingMixin


class PageContentManager(LoggingMixin):
    def __init__(
        self,
        page_id: str,
        client: NotionClient,
        block_registry: Optional[BlockElementRegistry] = None,
    ):
        self.page_id = page_id
        self._client = client
        self._markdown_to_notion_converter = MarkdownToNotionConverter(
            block_registry=block_registry
        )
        self._notion_to_markdown_converter = NotionToMarkdownConverter(
            block_registry=block_registry
        )

    async def append_markdown(self, markdown_text: str) -> str:
        blocks = self._markdown_to_notion_converter.convert(markdown_text)
        result = await self._client.patch(
            f"blocks/{self.page_id}/children", {"children": blocks}
        )
        return (
            "Successfully added text to the page." if result else "Failed to add text."
        )

    async def clear(self) -> str:
        blocks = await self._client.get(f"blocks/{self.page_id}/children")
        if not blocks:
            return "No content to delete."

        results = blocks.get("results", [])
        if not results:
            return "No content to delete."

        deleted = 0
        for b in results:
            if await self._client.delete(f"blocks/{b['id']}"):
                deleted += 1

        return f"Deleted {deleted}/{len(results)} blocks."

    # Methods from PageContentReader
    async def get_blocks(self) -> List[Dict[str, Any]]:
        result = await self._client.get(f"blocks/{self.page_id}/children")
        if not result:
            self.logger.error("Error retrieving page content: %s", result.error)
            return []
        return result.data.get("results", [])

    async def get_block_children(self, block_id: str) -> List[Dict[str, Any]]:
        result = await self._client.get(f"blocks/{block_id}/children")
        if not result:
            self.logger.error("Error retrieving block children: %s", result.error)
            return []
        return result.data.get("results", [])

    async def get_page_blocks_with_children(self) -> List[Dict[str, Any]]:
        blocks = await self.get_blocks()
        for block in blocks:
            if block.get("has_children"):
                block_id = block.get("id")
                if block_id:
                    children = await self.get_block_children(block_id)
                    if children:
                        block["children"] = children
        return blocks

    async def get_text(self) -> str:
        blocks = await self.get_page_blocks_with_children()
        return self._notion_to_markdown_converter.convert(blocks)



import re
from typing import Dict, Any, Optional, List, Tuple, Callable

from notionary.core.converters.elements.notion_block_element import NotionBlockElement


class ToggleElement(NotionBlockElement):
    """
    Handles conversion between Markdown toggle blocks and Notion toggle blocks.

    Markdown toggle syntax:
    +++ Toggle title
       Indented content that belongs to the toggle
       More indented content

    Non-indented content marks the end of the toggle block.
    """

    TOGGLE_PATTERN = re.compile(r"^[+]{3}\s+(.+)$")

    INDENT_PATTERN = re.compile(r"^(\s{2,}|\t+)(.+)$")

    @staticmethod
    def match_markdown(text: str) -> bool:
        """Check if text is a markdown toggle."""
        return bool(ToggleElement.TOGGLE_PATTERN.match(text.strip()))

    @staticmethod
    def match_notion(block: Dict[str, Any]) -> bool:
        """Check if block is a Notion toggle."""
        return block.get("type") == "toggle"

    @staticmethod
    def markdown_to_notion(text: str) -> Optional[Dict[str, Any]]:
        """Convert markdown toggle to Notion toggle block.

        Note: This method only converts the toggle title line.
        The nested content needs to be processed separately.
        """
        toggle_match = ToggleElement.TOGGLE_PATTERN.match(text.strip())
        if not toggle_match:
            return None

        # Extract content
        title = toggle_match.group(1)

        return {
            "type": "toggle",
            "toggle": {
                "rich_text": [{"type": "text", "text": {"content": title}}],
                "color": "default",
                "children": [],  # Will be populated with nested content
            },
        }

    @staticmethod
    def extract_nested_content(
        lines: List[str], start_index: int
    ) -> Tuple[List[str], int]:
        """
        Extract the nested content of a toggle element.

        Args:
            lines: All lines of text
            start_index: Starting index to look for nested content

        Returns:
            Tuple of (nested_content_lines, next_line_index)
        """
        nested_content = []
        current_index = start_index

        while current_index < len(lines):
            line = lines[current_index]

            # Empty line is still part of toggle content
            if not line.strip():
                nested_content.append("")
                current_index += 1
                continue

            # Check if line is indented (part of toggle content)
            if line.startswith("  ") or line.startswith("\t"):
                # Extract content with indentation removed
                content_line = ToggleElement._remove_indentation(line)
                nested_content.append(content_line)
                current_index += 1
                continue

            # Non-indented, non-empty line marks the end of toggle content
            break

        return nested_content, current_index

    @staticmethod
    def _remove_indentation(line: str) -> str:
        """Remove indentation from a line, handling both spaces and tabs."""
        if line.startswith("\t"):
            return line[1:]
        else:
            # Find number of leading spaces
            leading_spaces = len(line) - len(line.lstrip(" "))
            # Remove at least 2 spaces, but not more than what's there
            return line[min(2, leading_spaces) :]

    @staticmethod
    def notion_to_markdown(block: Dict[str, Any]) -> Optional[str]:
        """Convert Notion toggle block to markdown toggle."""
        if block.get("type") != "toggle":
            return None

        toggle_data = block.get("toggle", {})

        # Extract title from rich_text
        title = ToggleElement._extract_text_content(toggle_data.get("rich_text", []))

        # Create the toggle line
        toggle_line = f"+++ {title}"

        # Process children if any
        children = toggle_data.get("children", [])
        if children:
            child_lines = []
            for child_block in children:
                # This would need to be handled by a full converter that can dispatch
                # to the appropriate element type for each child block
                child_markdown = "  [Nested content]"  # Placeholder
                child_lines.append(f"  {child_markdown}")

            return toggle_line + "\n" + "\n".join(child_lines)

        return toggle_line

    @staticmethod
    def is_multiline() -> bool:
        """Toggle blocks can span multiple lines due to their nested content."""
        return True

    @staticmethod
    def _extract_text_content(rich_text: List[Dict[str, Any]]) -> str:
        """Extract plain text content from Notion rich_text elements."""
        result = ""
        for text_obj in rich_text:
            if text_obj.get("type") == "text":
                result += text_obj.get("text", {}).get("content", "")
            elif "plain_text" in text_obj:
                result += text_obj.get("plain_text", "")
        return result

    @classmethod
    def find_matches(
        cls, text: str, process_nested_content: Callable = None
    ) -> List[Tuple[int, int, Dict[str, Any]]]:
        """
        Find all toggle elements in the text and process them.

        Args:
            text: The text to search in
            process_nested_content: Optional callback function to process nested content
                It should accept a string and return a list of Notion blocks

        Returns:
            List of (start_pos, end_pos, block) tuples
        """
        if not text:
            return []

        toggle_blocks = []
        lines = text.split("\n")

        i = 0
        while i < len(lines):
            line = lines[i]

            # Check if line is a toggle
            if not cls.match_markdown(line):
                i += 1
                continue

            start_pos = 0
            for j in range(i):
                start_pos += len(lines[j]) + 1

            toggle_block = cls.markdown_to_notion(line)
            if not toggle_block:
                i += 1
                continue

            # Extract nested content
            nested_content, next_index = cls.extract_nested_content(lines, i + 1)

            # Calculate ending position
            end_pos = start_pos + len(line) + sum(len(l) + 1 for l in nested_content)

            if nested_content and process_nested_content:
                nested_text = "\n".join(nested_content)
                nested_blocks = process_nested_content(nested_text)
                if nested_blocks:
                    toggle_block["toggle"]["children"] = nested_blocks

            toggle_blocks.append((start_pos, end_pos, toggle_block))

            i = next_index

        return toggle_blocks



from typing import Any, AsyncGenerator, Dict, List, Optional
from notionary.core.database.notion_database_schema import NotionDatabaseSchema
from notionary.core.page.notion_page_manager import NotionPageManager


class DatabaseQueryService:
    """Service für Datenbankabfragen und Iterations"""

    def __init__(self, schema: NotionDatabaseSchema):
        self._schema = schema

    async def get_pages(
        self,
        database_id: str,
        limit: int = 100,
        filter_conditions: Optional[Dict[str, Any]] = None,
        sorts: Optional[List[Dict[str, Any]]] = None,
    ) -> List[NotionPageManager]:
        """
        Get all pages from the database.

        Args:
            database_id: The database ID to query
            limit: Maximum number of pages to retrieve
            filter_conditions: Optional filter to apply to the database query
            sorts: Optional sort instructions for the database query

        Returns:
            List of NotionPageManager instances for each page
        """
        pages: List[NotionPageManager] = []
        count = 0

        async for page in self.iter_pages(
            database_id,
            page_size=min(limit, 100),
            filter_conditions=filter_conditions,
            sorts=sorts,
        ):
            pages.append(page)
            count += 1

            if count >= limit:
                break

        return pages

    async def iter_pages(
        self,
        database_id: str,
        page_size: int = 100,
        filter_conditions: Optional[Dict[str, Any]] = None,
        sorts: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncGenerator[NotionPageManager, None]:
        """
        Asynchronous generator that yields pages from the database.

        Args:
            database_id: The database ID to query
            page_size: Number of pages to fetch per request
            filter_conditions: Optional filter to apply to the database query
            sorts: Optional sort instructions for the database query

        Yields:
            NotionPageManager instances for each page
        """
        async for page_manager in self._schema.iter_database_pages(
            database_id=database_id,
            page_size=page_size,
            filter_conditions=filter_conditions,
            sorts=sorts,
        ):
            yield page_manager



from typing import Dict, Any, List, Optional

from notionary.core.converters.registry.block_element_registry import (
    BlockElementRegistry,
)
from notionary.core.converters.registry.block_element_registry_builder import (
    BlockElementRegistryBuilder,
)


class NotionToMarkdownConverter:
    """Converts Notion blocks to Markdown text."""

    def __init__(self, block_registry: Optional[BlockElementRegistry] = None):
        """
        Initialize the MarkdownToNotionConverter.

        Args:
            block_registry: Optional registry of Notion block elements
        """
        self._block_registry = (
            block_registry or BlockElementRegistryBuilder().create_standard_registry()
        )

    def convert(self, blocks: List[Dict[str, Any]]) -> str:
        """
        Convert Notion blocks to Markdown text.

        Args:
            blocks: List of Notion blocks

        Returns:
            Markdown text
        """
        if not blocks:
            return ""

        markdown_parts = []

        for block in blocks:
            markdown = self._block_registry.notion_to_markdown(block)
            if markdown:
                markdown_parts.append(markdown)

        return "\n\n".join(markdown_parts)



from typing import Dict, Any, List, Tuple
import re


class TextInlineFormatter:
    """
    Handles conversion between Markdown inline formatting and Notion rich text elements.

    Supports various formatting options:
    - Bold: **text**
    - Italic: *text* or _text_
    - Underline: __text__
    - Strikethrough: ~~text~~
    - Code: `text`
    - Links: [text](url)
    - Highlights: ==text== (default yellow) or ==color:text== (custom color)
    """

    # Format patterns for matching Markdown formatting
    FORMAT_PATTERNS = [
        (r"\*\*(.+?)\*\*", {"bold": True}),
        (r"\*(.+?)\*", {"italic": True}),
        (r"_(.+?)_", {"italic": True}),
        (r"__(.+?)__", {"underline": True}),
        (r"~~(.+?)~~", {"strikethrough": True}),
        (r"`(.+?)`", {"code": True}),
        (r"\[(.+?)\]\((.+?)\)", {"link": True}),
        (r"==([a-z_]+):(.+?)==", {"highlight": True}),
        (r"==(.+?)==", {"highlight_default": True}),
    ]

    # Valid colors for highlighting
    VALID_COLORS = [
        "default",
        "gray",
        "brown",
        "orange",
        "yellow",
        "green",
        "blue",
        "purple",
        "pink",
        "red",
        "gray_background",
        "brown_background",
        "orange_background",
        "yellow_background",
        "green_background",
        "blue_background",
        "purple_background",
        "pink_background",
        "red_background",
    ]

    @classmethod
    def parse_inline_formatting(cls, text: str) -> List[Dict[str, Any]]:
        """
        Parse inline text formatting into Notion rich_text format.

        Args:
            text: Markdown text with inline formatting

        Returns:
            List of Notion rich_text objects
        """
        if not text:
            return []

        return cls._split_text_into_segments(text, cls.FORMAT_PATTERNS)

    @classmethod
    def _split_text_into_segments(
        cls, text: str, format_patterns: List[Tuple]
    ) -> List[Dict[str, Any]]:
        """
        Split text into segments by formatting markers and convert to Notion rich_text format.

        Args:
            text: Text to split
            format_patterns: List of (regex pattern, formatting dict) tuples

        Returns:
            List of Notion rich_text objects
        """
        segments = []
        remaining_text = text

        while remaining_text:
            earliest_match = None
            earliest_format = None
            earliest_pos = len(remaining_text)

            # Find the earliest formatting marker
            for pattern, formatting in format_patterns:
                match = re.search(pattern, remaining_text)
                if match and match.start() < earliest_pos:
                    earliest_match = match
                    earliest_format = formatting
                    earliest_pos = match.start()

            if earliest_match is None:
                if remaining_text:
                    segments.append(cls._create_text_element(remaining_text, {}))
                break

            if earliest_pos > 0:
                segments.append(
                    cls._create_text_element(remaining_text[:earliest_pos], {})
                )

            if "highlight" in earliest_format:
                color = earliest_match.group(1)
                content = earliest_match.group(2)

                if color not in cls.VALID_COLORS:
                    if not color.endswith("_background"):
                        color = f"{color}_background"

                    if color not in cls.VALID_COLORS:
                        color = "yellow_background"

                segments.append(cls._create_text_element(content, {"color": color}))

            elif "highlight_default" in earliest_format:
                content = earliest_match.group(1)
                segments.append(
                    cls._create_text_element(content, {"color": "yellow_background"})
                )

            elif "link" in earliest_format:
                content = earliest_match.group(1)
                url = earliest_match.group(2)
                segments.append(cls._create_link_element(content, url))

            else:
                content = earliest_match.group(1)
                segments.append(cls._create_text_element(content, earliest_format))

            # Move past the processed segment
            remaining_text = remaining_text[
                earliest_pos + len(earliest_match.group(0)) :
            ]

        return segments

    @classmethod
    def _create_text_element(
        cls, text: str, formatting: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Create a Notion text element with formatting.

        Args:
            text: The text content
            formatting: Dictionary of formatting options

        Returns:
            Notion rich_text element
        """
        annotations = cls._default_annotations()

        # Apply formatting
        for key, value in formatting.items():
            if key == "color":
                annotations["color"] = value
            elif key in annotations:
                annotations[key] = value

        return {
            "type": "text",
            "text": {"content": text},
            "annotations": annotations,
            "plain_text": text,
        }

    @classmethod
    def _create_link_element(cls, text: str, url: str) -> Dict[str, Any]:
        """
        Create a Notion link element.

        Args:
            text: The link text
            url: The URL

        Returns:
            Notion rich_text element with link
        """
        return {
            "type": "text",
            "text": {"content": text, "link": {"url": url}},
            "annotations": cls._default_annotations(),
            "plain_text": text,
        }

    @classmethod
    def extract_text_with_formatting(cls, rich_text: List[Dict[str, Any]]) -> str:
        """
        Convert Notion rich_text elements back to Markdown formatted text.

        Args:
            rich_text: List of Notion rich_text elements

        Returns:
            Markdown formatted text
        """
        formatted_parts = []

        for text_obj in rich_text:
            # Fallback: If plain_text is missing, use text['content']
            content = text_obj.get("plain_text")
            if content is None:
                content = text_obj.get("text", {}).get("content", "")

            annotations = text_obj.get("annotations", {})

            if annotations.get("code", False):
                content = f"`{content}`"
            if annotations.get("strikethrough", False):
                content = f"~~{content}~~"
            if annotations.get("underline", False):
                content = f"__{content}__"
            if annotations.get("italic", False):
                content = f"*{content}*"
            if annotations.get("bold", False):
                content = f"**{content}**"

            color = annotations.get("color", "default")
            if color != "default":
                content = f"=={color.replace('_background', '')}:{content}=="

            text_data = text_obj.get("text", {})
            link_data = text_data.get("link")
            if link_data:
                url = link_data.get("url", "")
                content = f"[{content}]({url})"

            formatted_parts.append(content)

        return "".join(formatted_parts)

    @classmethod
    def _default_annotations(cls) -> Dict[str, bool]:
        """
        Create default annotations object.

        Returns:
            Default Notion text annotations
        """
        return {
            "bold": False,
            "italic": False,
            "strikethrough": False,
            "underline": False,
            "code": False,
            "color": "default",
        }

    @classmethod
    def get_llm_prompt_content(cls) -> Dict[str, Any]:
        """
        Returns information about inline text formatting capabilities for LLM prompts.

        This method provides documentation about supported inline formatting options
        that can be used across all block elements.

        Returns:
            A dictionary with descriptions, syntax examples, and usage guidelines
        """
        return {
            "description": "Standard Markdown formatting is supported in all text blocks. Additionally, a custom highlight syntax is available for emphasizing important information. To create vertical spacing between elements, use the special spacer tag.",
            "syntax": [
                "**text** - Bold text",
                "*text* or _text_ - Italic text",
                "__text__ - Underlined text",
                "~~text~~ - Strikethrough text",
                "`text` - Inline code",
                "[text](url) - Link",
                "==text== - Default highlight (yellow background)",
                "==color:text== - Colored highlight (e.g., ==red:warning==)",
                "<!-- spacer --> - Creates vertical spacing between elements",
            ],
            "examples": [
                "This is a **bold** statement with some *italic* words.",
                "This feature is ~~deprecated~~ as of version 2.0.",
                "Edit the `config.json` file to configure settings.",
                "Check our [documentation](https://docs.example.com) for more details.",
                "==This is an important note== that you should remember.",
                "==red:Warning:== This action cannot be undone.",
                "==blue:Note:== Common colors include red, blue, green, yellow, purple.",
                "First paragraph content.\n\n<!-- spacer -->\n\nSecond paragraph with additional spacing above.",
            ],
            "highlight_usage": "The highlight syntax (==text== and ==color:text==) should be used to emphasize important information, warnings, notes, or other content that needs to stand out. This is particularly useful for making content more scannable at a glance.",
            "spacer_usage": "Use the <!-- spacer --> tag on its own line to create additional vertical spacing between elements. This is useful for improving readability by visually separating sections of content. Multiple spacer tags can be used for greater spacing.",
        }



from typing import Any, Dict, Optional
from notionary.core.notion_client import NotionClient
from notionary.util.logging_mixin import LoggingMixin


class MetadataEditor(LoggingMixin):
    def __init__(self, page_id: str, client: NotionClient):
        self.page_id = page_id
        self._client = client

    async def set_title(self, title: str) -> Optional[Dict[str, Any]]:
        return await self._client.patch(
            f"pages/{self.page_id}",
            {
                "properties": {
                    "title": {"title": [{"type": "text", "text": {"content": title}}]}
                }
            },
        )

    async def set_icon(
        self, emoji: Optional[str] = None, external_url: Optional[str] = None
    ) -> Optional[Dict[str, Any]]:
        if emoji:
            icon = {"type": "emoji", "emoji": emoji}
        elif external_url:
            icon = {"type": "external", "external": {"url": external_url}}
        else:
            return None

        return await self._client.patch(f"pages/{self.page_id}", {"icon": icon})

    async def set_cover(self, external_url: str) -> Optional[Dict[str, Any]]:
        return await self._client.patch(
            f"pages/{self.page_id}",
            {"cover": {"type": "external", "external": {"url": external_url}}},
        )
